import gc
import re
import os
from collections import Counter
from dataclasses import dataclass

import torch
import numpy as np
from tqdm import tqdm

from remora.refine_signal_map import SigMapRefiner
from remora import constants, log, RemoraError, util, encoded_kmers

LOGGER = log.get_logger()

DATASET_VERSION = 2
MISMATCH_ARRS = {
    0: np.array([1, 2, 3]),
    1: np.array([0, 2, 3]),
    2: np.array([0, 1, 3]),
    3: np.array([0, 1, 2]),
}
# CIGAR operations which correspond to query and reference sequence
MATCH_OPS = np.array(
    [True, False, False, False, False, False, False, True, True]
)
QUERY_OPS = np.array([True, True, False, False, True, False, False, True, True])
REF_OPS = np.array([True, False, True, True, False, False, False, True, True])
CIGAR_CODES = ["M", "I", "D", "N", "S", "H", "P", "=", "X"]
CODE_TO_OP = {
    "M": 0,
    "I": 1,
    "D": 2,
    "N": 3,
    "S": 4,
    "H": 5,
    "P": 6,
    "=": 7,
    "X": 8,
}
CIGAR_STRING_PATTERN = re.compile(r"(\d+)" + f"([{''.join(CIGAR_CODES)}])")


def cigartuples_from_string(cigarstring):
    """
    Returns pysam-style list of (op, count) tuples from a cigarstring.
    """
    return [
        (CODE_TO_OP[m.group(2)], int(m.group(1)))
        for m in re.finditer(CIGAR_STRING_PATTERN, cigarstring)
    ]


def map_ref_to_signal(*, query_to_signal, ref_to_query_knots):
    """Compute interpolated mapping from reference, through query alignment to
    signal coordinates

    Args:
        query_to_signal (np.array): Query to signal coordinate mapping
        ref_to_query_knots (np.array): Reference to query coordinate mapping
    """
    return np.floor(
        np.interp(
            ref_to_query_knots,
            np.arange(query_to_signal.size),
            query_to_signal,
        )
    ).astype(int)


def make_sequence_coordinate_mapping(cigar):
    """Maps an element in reference to every element in basecalls using
    alignment in `cigar`.

    Args:
        cigar (list): "cigartuples" representing alignment

    Returns:
        array shape (ref_len,). [x_0, x_1, ..., x_(ref_len)]
            such that read_seq[x_i] <> ref_seq[i]. Note that ref_len is derived
            from the cigar input.
    """
    ops, lens = map(np.array, zip(*cigar))
    assert ops.min() >= 0 and ops.max() <= 8, "invalid cigar op(s)"
    assert lens.min() >= 0, "cigar lengths may not be negative"

    is_match = MATCH_OPS[ops]
    match_counts = lens[is_match]
    offsets = np.array([match_counts, np.ones_like(match_counts)])

    # TODO remove knots around ambiguous indels (e.g. left justified HPs)
    # note this requires the ref and query sequences
    ref_knots = np.cumsum(np.where(REF_OPS[ops], lens, 0))
    ref_knots = np.concatenate(
        [[0], (ref_knots[is_match] - offsets).T.flatten(), [ref_knots[-1]]]
    )
    query_knots = np.cumsum(np.where(QUERY_OPS[ops], lens, 0))
    query_knots = np.concatenate(
        [[0], (query_knots[is_match] - offsets).T.flatten(), [query_knots[-1]]]
    )
    knots = np.interp(np.arange(ref_knots[-1] + 1), ref_knots, query_knots)

    return knots


def compute_ref_to_signal(query_to_signal, cigar):
    ref_to_read_knots = make_sequence_coordinate_mapping(cigar)
    assert query_to_signal.size == ref_to_read_knots[-1].astype(int) + 1, (
        "discordant implied basecall lengths: move_table:"
        f"{query_to_signal.size} cigar:{ref_to_read_knots[-1]}"
    )
    return map_ref_to_signal(
        query_to_signal=query_to_signal, ref_to_query_knots=ref_to_read_knots
    )


@dataclass
class RemoraRead:
    """Object to hold information about a read relevant to Remora training and
    inference.

    Args:
        dacs (np.ndarray): Unnormalized DAC signal. `dacs` should be reversed
            already for reverse_signal data types.
        shift (float): Shift from dac to normalized signal. via formula:
            norm = (dac - shift) / scale
        scale (float): Scale from dac to normalized signal
        seq_to_sig_map (np.ndarray): Position within signal array assigned to
            each base in seq
        int_seq (np.ndarray): Encoded sequence for training/validation.
            See remora.util.seq_to_int
        str_seq (str): String sequence for training/validation. Ignored if
            int_seq is provided.
        read_id (str): Read identifier
        labels (np.ndarray): Output label for each base in read
        focus_bases (np.ndarray): Sites from read to produce calls
        batches (list): List of batches from RemoraDataset

    Note: Must provide either int_seq or str_seq. If str_seq is provided
    int_seq will be derived on init.
    """

    dacs: np.ndarray
    shift: float
    scale: float
    seq_to_sig_map: np.ndarray
    int_seq: np.ndarray = None
    str_seq: str = None
    read_id: str = None
    labels: np.ndarray = None
    focus_bases: np.ndarray = None
    batches: list = None

    def __post_init__(self):
        if self.int_seq is None:
            if self.str_seq is None:
                raise RemoraError(
                    "Must provide sequence to initialize RemoraRead"
                )
            # if int_seq is not set, set from str_seq provided
            self.int_seq = util.seq_to_int(self.str_seq)
        else:
            # set str_seq from int_seq to ensure the sequences match
            self.str_seq = util.int_to_seq(self.int_seq)
        self._sig = None
        self._dwells = None
        self._sig_cumsum = None
        self._base_levels = None

    @classmethod
    def test_read(cls, nbases=20, signal_per_base=10):
        """Spoofed read for testing"""
        return cls(
            np.zeros(nbases * signal_per_base),
            0.0,
            1.0,
            np.arange(nbases * signal_per_base + 1, step=signal_per_base),
            np.arange(nbases) % 4,
            "test_read",
            np.zeros(nbases, dtype=np.int64),
        )

    @property
    def sig(self):
        if self._sig is None:
            self._sig = (self.dacs - self.shift) / self.scale
        return self._sig

    @property
    def sig_cumsum(self):
        if self._sig_cumsum is None:
            self._sig_cumsum = np.empty(self.sig.size + 1)
            self._sig_cumsum[0] = 0
            self._sig_cumsum[1:] = np.cumsum(self.sig)
        return self._sig_cumsum

    @property
    def dwells(self):
        if self._dwells is None:
            self._dwells = np.diff(self.seq_to_sig_map)
        return self._dwells

    @property
    def base_levels(self):
        if self._base_levels is None:
            with np.errstate(invalid="ignore"):
                self._base_levels = (
                    np.diff(self.sig_cumsum[self.seq_to_sig_map]) / self.dwells
                )
        return self._base_levels

    def check(self):
        if self.seq_to_sig_map.size != self.int_seq.size + 1:
            LOGGER.debug(
                "Invalid read: seq and mapping mismatch "
                f"{self.seq_to_sig_map.size} != {self.int_seq.size + 1}"
            )
            raise RemoraError(
                f"Invalid read: seq ({self.int_seq.size}) and mapping "
                f"({self.seq_to_sig_map.size}) sizes incompatible"
            )
        if self.seq_to_sig_map[0] != 0:
            LOGGER.debug(
                f"Invalid read {self.read_id} : invalid mapping start "
                f"{self.seq_to_sig_map[0]} != 0"
            )
            raise RemoraError("Invalid read: mapping start")
        if self.seq_to_sig_map[-1] != self.sig.size:
            LOGGER.debug(
                f"Invalid read {self.read_id} : invalid mapping end "
                f"{self.seq_to_sig_map[-1]} != {self.sig.size}"
            )
            raise RemoraError("Invalid read: mapping end")
        if self.int_seq.max() > 3:
            LOGGER.debug(f"Invalid read: Invalid base {self.int_seq.max()}")
            raise RemoraError("Invalid read: Invalid base")
        if self.int_seq.min() < -1:
            LOGGER.debug(f"Invalid read: Invalid base {self.int_seq.min()}")
            raise RemoraError("Invalid read: Invalid base")
        # TODO add more logic to these tests

    def copy(self):
        return RemoraRead(
            dacs=self.dacs.copy(),
            shift=self.shift,
            scale=self.scale,
            seq_to_sig_map=self.seq_to_sig_map,
            int_seq=None if self.int_seq is None else self.int_seq.copy(),
            str_seq=self.str_seq,
            read_id=self.read_id,
            labels=None if self.labels is None else self.labels.copy(),
            focus_bases=None
            if self.focus_bases is None
            else self.focus_bases.copy(),
        )

    def refine_signal_mapping(self, sig_map_refiner, check_read=False):
        if not sig_map_refiner.is_loaded:
            LOGGER.debug("no signal map refiner loaded skipping..")
            return
        if sig_map_refiner.do_rough_rescale:
            prev_shift, prev_scale = self.shift, self.scale
            self.shift, self.scale = sig_map_refiner.rough_rescale(
                self.shift,
                self.scale,
                self.seq_to_sig_map,
                self.int_seq,
                self.dacs,
            )
            self._sig = None
            self._sig_cumsum = None
            self._base_levels = None
        if sig_map_refiner.scale_iters >= 0:
            prev_shift, prev_scale = self.shift, self.scale
            prev_seq_to_sig_map = self.seq_to_sig_map.copy()
            try:
                (
                    self.seq_to_sig_map,
                    self.shift,
                    self.scale,
                ) = sig_map_refiner.refine_sig_map(
                    self.shift,
                    self.scale,
                    self.seq_to_sig_map,
                    self.int_seq,
                    self.dacs,
                )
            except IndexError as e:
                LOGGER.debug(f"refine_error {self.read_id} {e}")
            # reset computed values after refinement
            self._sig = None
            self._dwells = None
            self._sig_cumsum = None
            self._base_levels = None
            LOGGER.debug(f"refine_mapping_shift: {prev_shift} {self.shift}")
            LOGGER.debug(f"refine_mapping_scale: {prev_scale} {self.scale}")
            sig_map_diffs = self.seq_to_sig_map - prev_seq_to_sig_map
            LOGGER.debug(
                f"refine_mapping_median_adjust: {np.median(sig_map_diffs)} "
                f"{self.read_id}"
            )
        if check_read:
            self.check()

    def set_motif_focus_bases(self, motifs):
        """
        Mutates self. Sets self.focus_bases to all hits within self.int_seq.
        :param motifs: Iterable of util.Motifs
        """
        self.focus_bases = util.find_focus_bases_in_int_sequence(
            self.int_seq, motifs
        )

    def downsample_focus_bases(self, max_sites):
        if self.focus_bases is not None and self.focus_bases.size > max_sites:
            LOGGER.debug(
                f"selected {max_sites} focus bases from "
                f"{self.focus_bases.size} in read {self.read_id}"
            )
            self.focus_bases = np.random.choice(
                self.focus_bases,
                size=max_sites,
                replace=False,
            )

    def extract_chunk(
        self,
        focus_sig_idx,
        chunk_context,
        kmer_context_bases,
        label=-1,
        base_pred=False,
        read_focus_base=-1,
        check_chunk=False,
    ):
        chunk_len = sum(chunk_context)
        sig_start = focus_sig_idx - chunk_context[0]
        sig_end = focus_sig_idx + chunk_context[1]
        seq_to_sig_offset = 0
        if sig_start >= 0 and sig_end <= self.sig.size:
            # chunk boundaries are within read signal
            chunk_sig = self.sig[sig_start:sig_end].copy()
        else:
            # if signal is not available for full chunk pad with zeros
            chunk_sig = np.zeros(chunk_len, dtype=np.float32)
            fill_st = 0
            fill_en = chunk_len
            if sig_start < 0:
                fill_st = -sig_start
                # record offset value by which to shift seq_to_sig_map
                seq_to_sig_offset = -sig_start
                sig_start = 0
            if sig_end > self.sig.size:
                fill_en = self.sig.size - sig_start
                sig_end = self.sig.size
            chunk_sig[fill_st:fill_en] = self.sig[sig_start:sig_end]

        seq_start = (
            np.searchsorted(self.seq_to_sig_map, sig_start, side="right") - 1
        )
        seq_end = np.searchsorted(self.seq_to_sig_map, sig_end, side="left")

        # extract/compute sequence to signal mapping for this chunk
        chunk_seq_to_sig = self.seq_to_sig_map[seq_start : seq_end + 1].copy()
        # shift mapping relative to the chunk
        chunk_seq_to_sig -= sig_start - seq_to_sig_offset
        # set chunk ends to chunk boundaries
        chunk_seq_to_sig[0] = 0
        chunk_seq_to_sig[-1] = chunk_len
        chunk_seq_to_sig = chunk_seq_to_sig.astype(np.int32)

        # extract context sequence
        kmer_before_bases, kmer_after_bases = kmer_context_bases
        if (
            seq_start >= kmer_before_bases
            and seq_end + kmer_after_bases <= self.int_seq.size
        ):
            chunk_seq = self.int_seq[
                seq_start - kmer_before_bases : seq_end + kmer_after_bases
            ]
        else:
            chunk_seq = np.full(
                seq_end - seq_start + sum(kmer_context_bases),
                -1,
                dtype=np.byte,
            )
            fill_st = 0
            fill_en = seq_end - seq_start + sum(kmer_context_bases)
            chunk_seq_st = seq_start - kmer_before_bases
            chunk_seq_en = seq_end + kmer_after_bases
            if seq_start < kmer_before_bases:
                fill_st = kmer_before_bases - seq_start
                chunk_seq_st = 0
            if seq_end + kmer_after_bases > self.int_seq.size:
                fill_en -= seq_end + kmer_after_bases - self.int_seq.size
                chunk_seq_en = self.int_seq.size
            chunk_seq[fill_st:fill_en] = self.int_seq[chunk_seq_st:chunk_seq_en]
        chunk = Chunk(
            signal=chunk_sig,
            seq_w_context=chunk_seq,
            seq_to_sig_map=chunk_seq_to_sig,
            kmer_context_bases=kmer_context_bases,
            chunk_sig_focus_idx=focus_sig_idx - sig_start,
            chunk_focus_base=read_focus_base - seq_start,
            read_focus_base=read_focus_base,
            read_id=self.read_id,
            label=label,
        )
        if check_chunk:
            chunk.check()
        if base_pred:
            chunk.mask_focus_base()
        return chunk

    def iter_chunks(
        self,
        chunk_context,
        kmer_context_bases,
        base_pred=False,
        base_start_justify=False,
        offset=0,
        check_chunks=False,
    ):
        for focus_base in self.focus_bases:
            label = -1 if self.labels is None else self.labels[focus_base]
            # add offset and ensure not out of bounds
            focus_base = max(
                min(focus_base + offset, self.seq_to_sig_map.size - 2), 0
            )
            if base_start_justify:
                focus_sig_idx = self.seq_to_sig_map[focus_base]
            else:
                # compute position at center of central base
                focus_sig_idx = (
                    self.seq_to_sig_map[focus_base]
                    + self.seq_to_sig_map[focus_base + 1]
                ) // 2
            try:
                yield self.extract_chunk(
                    focus_sig_idx,
                    chunk_context,
                    kmer_context_bases,
                    label=label,
                    base_pred=base_pred,
                    read_focus_base=focus_base,
                    check_chunk=check_chunks,
                )
            except RemoraError as e:
                LOGGER.debug(f"FAILED_CHUNK_CHECK {e}")
            except Exception as e:
                LOGGER.debug(f"FAILED_CHUNK_EXTRACT {e}")

    def prepare_batches(self, model_metadata, batch_size):
        """Prepare batches containing chunks from this read

        Args:
            model_metadata: Inference model metadata
            batch_size (int): Number of chunks to call per-batch
        """
        self.batches = []
        self.refine_signal_mapping(model_metadata["sig_map_refiner"])
        chunks = list(
            self.iter_chunks(
                model_metadata["chunk_context"],
                model_metadata["kmer_context_bases"],
                model_metadata["base_pred"],
                model_metadata["base_start_justify"],
                model_metadata["offset"],
            )
        )
        if len(chunks) == 0:
            return
        dataset = RemoraDataset.allocate_empty_chunks(
            num_chunks=len(chunks),
            chunk_context=model_metadata["chunk_context"],
            max_seq_len=max(c.seq_len for c in chunks),
            kmer_context_bases=model_metadata["kmer_context_bases"],
            base_pred=model_metadata["base_pred"],
            mod_bases=model_metadata["mod_bases"],
            mod_long_names=model_metadata["mod_long_names"],
            batch_size=batch_size,
            shuffle_on_iter=False,
            drop_last=False,
        )
        for chunk in chunks:
            dataset.add_chunk(chunk)
        dataset.set_nbatches()

        bb, ab = model_metadata["kmer_context_bases"]
        for (sigs, seqs, seq_maps, seq_lens), labels, (_, read_pos) in dataset:
            enc_kmers = encoded_kmers.compute_encoded_kmer_batch(
                bb, ab, seqs, seq_maps, seq_lens
            )
            self.batches.append((sigs, enc_kmers, labels, read_pos))

    def run_model(self, model):
        """Call modified bases on a read.

        Args:
            model: Compiled inference model (see remora.model_util.load_model)

        Returns:
            3-tuple containing:
              1. Modified base predictions (dim: num_calls, num_mods + 1)
              2. Labels for each base (-1 if labels not provided)
              3. List of positions within the read
        """
        device = next(model.parameters()).device
        read_outputs, read_poss, read_labels = [], [], []
        for sigs, enc_kmers, labels, read_pos in self.batches:
            sigs = torch.from_numpy(sigs).to(device)
            enc_kmers = torch.from_numpy(enc_kmers).to(device)
            output = model(sigs, enc_kmers).detach().cpu().numpy()
            read_outputs.append(output)
            read_labels.append(labels)
            read_poss.append(read_pos)
        read_outputs = np.concatenate(read_outputs, axis=0)
        read_labels = np.concatenate(read_labels)
        read_poss = np.concatenate(read_poss)
        return read_outputs, read_labels, read_poss


@dataclass
class Chunk:
    """Chunk of signal and associated sequence for training/infering results in
    Remora. Chunks are selected either with a given signal or bases location
    from a read.

    Args:
        signal (np.array): Normalized signal
        seq_w_context (str): Integer encoded sequence including context basees
            for kmer extraction. Note that seq_to_sig_map only corresponds to
            the central sequence without the context sequence
        seq_to_sig_map (np.array): Array of length one more than sequence with
            values representing indices into signal for each base.
        kmer_context_bases (tuple): Number of context bases included before and
            after the chunk sequence
        chunk_sig_focus_idx (int): Index within signal array on which the chunk
            is focuesed for prediction. May be used in model architecture in the
            future.
        chunk_focus_base (int): Index within chunk sequence (without context
            bases) on which the chunk is focuesed for prediction.
        read_focus_base (int): Position within full read for validation purposes
        read_id (str): Read ID
        label (int): Integer label for training/validation.
    """

    signal: np.ndarray
    seq_w_context: np.ndarray
    seq_to_sig_map: np.ndarray
    kmer_context_bases: tuple
    chunk_sig_focus_idx: int
    chunk_focus_base: int
    read_focus_base: int
    read_id: str = None
    label: int = None
    _base_sig_lens: np.ndarray = None

    def mask_focus_base(self):
        self.seq_w_context[
            self.chunk_focus_base + self.kmer_context_bases[0]
        ] = -1

    def check(self):
        if self.signal.size <= 0:
            LOGGER.debug(
                f"FAILED_CHUNK: no_sig {self.read_id} {self.read_focus_base}"
            )
            raise RemoraError("No signal for chunk")
        if (
            self.seq_w_context.size - sum(self.kmer_context_bases)
            != self.seq_to_sig_map.size - 1
        ):
            LOGGER.debug(
                f"FAILED_CHUNK: map_len {self.read_id} {self.read_focus_base}"
            )
            raise RemoraError("Invalid sig to seq map length")
        if not np.all(np.diff(self.seq_to_sig_map) >= 0):
            LOGGER.debug(
                f"FAILED_CHUNK: not monotonic {self.read_id} "
                f"{self.seq_to_sig_map}"
            )
        if self.seq_to_sig_map[0] < 0:
            LOGGER.debug(
                f"FAILED_CHUNK: start<0 {self.read_id} {self.seq_to_sig_map[0]}"
            )
            raise RemoraError("Seq to sig map starts before 0")
        if self.seq_to_sig_map[-1] > self.signal.size:
            LOGGER.debug(
                f"FAILED_CHUNK: end>sig_len {self.read_id} "
                f"{self.seq_to_sig_map[-1]}"
            )
            raise RemoraError("Seq to sig map ends after signal")
        # TODO add more logic to these checks

    @property
    def kmer_len(self):
        return sum(self.kmer_context_bases) + 1

    @property
    def seq_len(self):
        return self.seq_w_context.size - sum(self.kmer_context_bases)

    @property
    def seq(self):
        return self.seq_w_context[
            self.kmer_context_bases[0] : self.kmer_context_bases[0]
            + self.seq_len
        ]

    @property
    def base_sig_lens(self):
        if self._base_sig_lens is None:
            self._base_sig_lens = np.diff(self.seq_to_sig_map)
        return self._base_sig_lens


@dataclass
class RemoraDataset:
    """Remora dataset

    Args:
        sig_tensor (torch.FloatTensor): Signal chunks (dims: nchunks, 1, ntime)
        seq_array (np.array): Variable width integer encoded sequence chunks
            dtype : np.byte
            dims  : nchunks, max_seq_len + sum(kmer_context_bases)
        seq_mappings (np.array): Mapping from seq positions to signal tensor
            Note only sequence without kmer_context_bases bases is included in
            mappings
            dtype : np.short
            dims  : nchunks, max_seq_len + 1
        seq_lens (np.array): Length of sequences (without kmer context bases)
            dtype : np.short
            dims  : nchunks
        labels: (np.array): Chunk labels
            dtype : np.int64
            dims  : nchunks
        read_ids (np.array): Read IDs
            dtype : "U36"
            dims  : nchunks
        read_focus_bases (np.array): Base position within the read. This may be
            the basecalled sequence or the read-oriented reference sequence
            position depending on the extraction method.
            dtype : "U36"
            dims  : nchunks
        nchunks (int): Current number of chunks. If None (default), will be set
            to sig_tensor.shape[0]. If set, only chunks up to this value are
            assumed to be valid.
        chunk_context (tuple): 2-tuple containing the number of signal points
            before and after the central position.
        max_seq_len (int): Maximum sequence length of a chunk (used to set
            dimension for seq arrays.
        kmer_context_bases (tuple): 2-tuple containing the bases to include in
            the encoded k-mer presented as input.
        base_pred (bool): Are labels predicting base? Default is mod_bases.
        mod_bases (str): Modified base single letter codes represented by labels
        mod_long_names (list): Modified base long names represented by labels
        motifs (list): Tuples containing sequence motifs and relative position.
            E.g. ("N", 0) for all-contexts or ("CG", 0) for CG-contexts.
        reverse_signal (bool): Is nanopore signal 3' to 5' orientation?
            Primarily for directRNA
        batch_size (int): Size of batches to be produced
        shuffle_on_iter (bool): Shuffle data before each iteration over batches
        drop_last (bool): Drop the last batch of each iteration
        batch_label_props (np.array): Select proportion of each label
        sig_map_refiner (remora.refine_signal_map.SigMapRefiner): Signal
            mapping refiner
        base_start_justify (bool): Extract chunk centered on start of base
        offset (int): Extract chunk centered on base offset from base of
            interest
        drop_read_attrs (bool): Drop read id and read focus positions

    Yields:
        Batches of data for training or inference
    """

    # data attributes
    sig_tensor: np.ndarray
    seq_array: np.ndarray
    seq_mappings: np.ndarray
    seq_lens: np.ndarray
    labels: np.ndarray
    read_ids: np.ndarray
    read_focus_bases: np.ndarray
    nchunks: int = None

    # scalar metadata attributes
    chunk_context: tuple = constants.DEFAULT_CHUNK_CONTEXT
    max_seq_len: int = None
    kmer_context_bases: tuple = constants.DEFAULT_KMER_CONTEXT_BASES
    base_pred: bool = False
    mod_bases: str = ""
    mod_long_names: list = None
    motifs: list = None
    reverse_signal: bool = False

    # batch attributes (defaults set for training)
    batch_size: int = constants.DEFAULT_BATCH_SIZE
    shuffle_on_iter: bool = True
    drop_last: bool = True
    batch_label_props: np.ndarray = None

    # signal mapping refinement params
    sig_map_refiner: SigMapRefiner = None

    # chunk extraction attributes
    base_start_justify: bool = False
    offset: int = 0

    # extra attributes
    drop_read_attrs: bool = False

    def set_nbatches(self):
        self.nbatches, remainder = divmod(self.nchunks, self.batch_size)
        if not self.drop_last and remainder > 0:
            self.nbatches += 1

    def __post_init__(self):
        if self.max_seq_len is None:
            self.max_seq_len = self.seq_mappings.shape[1] - 1
        if self.nchunks is None:
            self.nchunks = self.sig_tensor.shape[0]
        elif self.nchunks > self.sig_tensor.shape[0]:
            raise RemoraError("More chunks indicated than provided.")
        if not self.base_pred and len(self.mod_bases) != len(
            self.mod_long_names
        ):
            raise RemoraError(
                f"mod_long_names ({self.mod_long_names}) must be same length "
                f"as mod_bases ({self.mod_bases})"
            )
        if self.motifs is not None:
            self.motifs = sorted(set(self.motifs))
        self.set_nbatches()
        self.shuffled = False
        if self.drop_read_attrs:
            self.read_ids = None
            self.read_focus_bases = None
            gc.collect()
        self._class_ordered_indices = None

    def add_chunk(self, chunk):
        if self.nchunks >= self.labels.size:
            raise RemoraError("Cannot add chunk to currently allocated tensors")
        if chunk.seq_len > self.max_seq_len:
            raise RemoraError("Chunk sequence too long to store")
        self.sig_tensor[self.nchunks, 0] = chunk.signal
        self.seq_array[
            self.nchunks, : chunk.seq_w_context.size
        ] = chunk.seq_w_context
        self.seq_mappings[
            self.nchunks, : chunk.seq_to_sig_map.size
        ] = chunk.seq_to_sig_map
        self.seq_lens[self.nchunks] = chunk.seq_len
        self.labels[self.nchunks] = chunk.label
        if not self.drop_read_attrs:
            self.read_ids[self.nchunks] = chunk.read_id
            self.read_focus_bases[self.nchunks] = chunk.read_focus_base
        self.nchunks += 1

    def add_batch(
        self, b_sig, b_seq, b_ss_map, b_seq_lens, b_labels, b_rids, b_rfbs
    ):
        batch_size = b_labels.size
        b_st, b_en = self.nchunks, self.nchunks + batch_size
        if self.nchunks + batch_size > self.labels.size:
            raise RemoraError("Cannot add batch to currently allocated tensors")
        # TODO check that applicable dims are compatible
        self.sig_tensor[b_st:b_en] = b_sig
        self.seq_array[b_st:b_en] = b_seq
        self.seq_mappings[b_st:b_en] = b_ss_map
        self.seq_lens[b_st:b_en] = b_seq_lens
        self.labels[b_st:b_en] = b_labels
        if not self.drop_read_attrs:
            self.read_ids[b_st:b_en] = b_rids
            self.read_focus_bases[b_st:b_en] = b_rfbs
        self.nchunks += batch_size

    def clip_chunks(self):
        if self.nchunks == self.sig_tensor.shape[0]:
            self.set_nbatches()
            return
        elif self.nchunks > self.sig_tensor.shape[0]:
            raise RemoraError("More chunks indicated than provided.")
        self.sig_tensor = self.sig_tensor[: self.nchunks]
        self.seq_array = self.seq_array[: self.nchunks]
        self.seq_mappings = self.seq_mappings[: self.nchunks]
        self.seq_lens = self.seq_lens[: self.nchunks]
        self.labels = self.labels[: self.nchunks]
        if not self.drop_read_attrs:
            self.read_ids = self.read_ids[: self.nchunks]
            self.read_focus_bases = self.read_focus_bases[: self.nchunks]
        # reset nbatches after clipping dataset
        self.set_nbatches()

    def trim_kmer_context_bases(self, new_kmer_context_bases):
        if new_kmer_context_bases is None:
            return
        if (
            new_kmer_context_bases[0] == self.kmer_context_bases[0]
            and new_kmer_context_bases[1] == self.kmer_context_bases[1]
        ):
            return
        if (
            new_kmer_context_bases[0] > self.kmer_context_bases[0]
            or new_kmer_context_bases[1] > self.kmer_context_bases[1]
        ):
            raise RemoraError(
                f"Cannot expand kmer context (prev:{self.kmer_context_bases} ; "
                f"requested_new:{new_kmer_context_bases})"
            )
        if new_kmer_context_bases[0] < self.kmer_context_bases[0]:
            seq_diff = self.kmer_context_bases[0] - new_kmer_context_bases[0]
            self.seq_array = np.ascontiguousarray(self.seq_array[:, seq_diff:])
        self.kmer_context_bases = new_kmer_context_bases

    def trim_chunk_context(self, new_chunk_context):
        if new_chunk_context is None:
            return
        if (
            new_chunk_context[0] == self.chunk_context[0]
            and new_chunk_context[1] == self.chunk_context[1]
        ):
            return
        if (
            new_chunk_context[0] > self.chunk_context[0]
            or new_chunk_context[1] > self.chunk_context[1]
        ):
            raise RemoraError(
                f"Cannot expand chunk_context (prev:{self.chunk_context} ; "
                f"requested_new:{new_chunk_context})"
            )
        raise NotImplementedError("Cannot currently trim chunk context.")

    def get_label_nchunks(self, target_nchunks):
        """Get the number of chunks for each label most closely matching
        specified batch_label_props,
        """
        if self.batch_label_props is None:
            raise RemoraError("Must set batch_label_props to get label nchunks")
        lab_nchunks = np.floor(target_nchunks * self.batch_label_props).astype(
            int
        )
        while lab_nchunks.sum() < target_nchunks:
            lab_nchunks[
                np.argmin(
                    (lab_nchunks / lab_nchunks.sum()) - self.batch_label_props
                )
            ] += 1
        return lab_nchunks

    def reorder_chunks(self, indices):
        """Re-order chunks via specified indices"""
        for attr in (
            "sig_tensor",
            "seq_array",
            "seq_mappings",
            "seq_lens",
            "labels",
        ):
            setattr(self, attr, getattr(self, attr)[indices])
            gc.collect()
        if not self.drop_read_attrs:
            self.read_ids = self.read_ids[indices]
            self.read_focus_bases = self.read_focus_bases[indices]
        gc.collect()

    def order_chunks_by_batch_labels(self):
        """Order chunks such that each slice of size self.batch_size includes
        the specified proportion of each label. Also store the indices
        cooresponding to each label to facilitate shuffling within each label
        while maintaining specified batch balance.

        Note that "extra" chunks are stored at the end of the dataset and will
        be shuffled into the used portion of the dataset at each call to
        shuffle_by_batch_labels.
        """
        class_indices = [
            np.where(self.labels == class_label)[0]
            for class_label in range(self.num_labels)
        ]
        batch_nchunks = self.get_label_nchunks(self.batch_size)
        class_nbatches = [
            ci.size // bnc for ci, bnc in zip(class_indices, batch_nchunks)
        ]
        # set to number of batches given smallest class size
        self.nbatches = min(class_nbatches)
        if self.nbatches < 1:
            raise RemoraError("Ordering by batch labels produced no batches")
        # set indices for each label within each batch
        idx_starts = np.concatenate([[0], np.cumsum(batch_nchunks)])
        class_extras = [
            ci.size - (bnc * self.nbatches)
            for ci, bnc in zip(class_indices, batch_nchunks)
        ]
        LOGGER.debug("Computing class order indices")
        # store indcies for each class to order all chunks and save to shuffle
        # within class each epoch
        self._class_ordered_indices = []
        extras_index = self.batch_size * self.nbatches
        for lab in range(self.num_labels):
            lab_b_rng = np.arange(idx_starts[lab], idx_starts[lab + 1])
            self._class_ordered_indices.append(
                np.concatenate(
                    [
                        lab_b_rng + (bn * self.batch_size)
                        for bn in range(self.nbatches)
                    ]
                    + [
                        np.arange(
                            extras_index, extras_index + class_extras[lab]
                        )
                    ]
                )
            )
            extras_index += class_extras[lab]
        global_ordered_indices = np.empty(self.nchunks, dtype=int)
        for ci, coi in zip(class_indices, self._class_ordered_indices):
            global_ordered_indices[coi] = ci
        gc.collect()
        LOGGER.debug("Reordering chunks")
        # order each tensor to batch ordering
        self.reorder_chunks(global_ordered_indices)

    def shuffle_by_batch_labels(self):
        if self._class_ordered_indices is None:
            raise RemoraError(
                "Must run order_chunks_by_batch_labels before "
                "shuffle_by_batch_labels"
            )
        global_ordered_indices = np.empty(self.nchunks, dtype=int)
        for coi in self._class_ordered_indices:
            global_ordered_indices[coi] = np.random.permutation(coi)
        self.reorder_chunks(global_ordered_indices)
        self.shuffled = True

    def shuffle(self):
        if not self.is_clipped:
            raise RemoraError("Cannot shuffle an unclipped dataset.")
        if self.batch_label_props is not None:
            raise RemoraError(
                "Cannot shuffle dataset with batch_label_props. "
                "Use shuffle_by_batch_labels"
            )
        shuf_idx = np.random.permutation(self.nchunks)
        self.reorder_chunks(shuf_idx)
        self.shuffled = True

    def clone_metadata(self):
        return {
            "chunk_context": self.chunk_context,
            "max_seq_len": self.max_seq_len,
            "kmer_context_bases": self.kmer_context_bases,
            "base_pred": self.base_pred,
            "mod_bases": self.mod_bases,
            "mod_long_names": self.mod_long_names,
            "motifs": self.motifs,
            "reverse_signal": self.reverse_signal,
            "batch_size": self.batch_size,
            "sig_map_refiner": self.sig_map_refiner,
            "offset": self.offset,
            "base_start_justify": self.base_start_justify,
        }

    def head(
        self,
        nchunks=None,
        prop=0.01,
        shuffle_on_iter=False,
        drop_last=False,
        stratified=False,
    ):
        if nchunks is None:
            nchunks = int(prop * self.nchunks)
        read_ids = read_focus_bases = None
        if stratified:
            LOGGER.debug("Performing stratified head")
            head_indices = []
            for class_label in range(self.num_labels):
                class_indices = np.where(self.labels == class_label)[0]
                if class_indices.size == 0:
                    continue
                cls_val_idx = int(class_indices.size * nchunks / self.nchunks)
                head_indices.append(class_indices[:cls_val_idx])

            head_indices = np.concatenate(head_indices)
            if not self.drop_read_attrs:
                read_ids = self.read_ids[head_indices].copy()
                read_focus_bases = self.read_focus_bases[head_indices].copy()
            return RemoraDataset(
                self.sig_tensor[head_indices].copy(),
                self.seq_array[head_indices].copy(),
                self.seq_mappings[head_indices].copy(),
                self.seq_lens[head_indices].copy(),
                self.labels[head_indices].copy(),
                read_ids,
                read_focus_bases,
                shuffle_on_iter=shuffle_on_iter,
                drop_last=drop_last,
                drop_read_attrs=self.drop_read_attrs,
                **self.clone_metadata(),
            )

        if not self.drop_read_attrs:
            read_ids = self.read_ids[:nchunks].copy()
            read_focus_bases = self.read_focus_bases[:nchunks].copy()
        return RemoraDataset(
            self.sig_tensor[:nchunks].copy(),
            self.seq_array[:nchunks].copy(),
            self.seq_mappings[:nchunks].copy(),
            self.seq_lens[:nchunks].copy(),
            self.labels[:nchunks].copy(),
            read_ids,
            read_focus_bases,
            shuffle_on_iter=shuffle_on_iter,
            drop_last=drop_last,
            drop_read_attrs=self.drop_read_attrs,
            **self.clone_metadata(),
        )

    def copy(self):
        read_ids = read_focus_bases = None
        if not self.drop_read_attrs:
            read_ids = self.read_ids.copy()
            read_focus_bases = self.read_focus_bases.copy()
        return RemoraDataset(
            self.sig_tensor.copy(),
            self.seq_array.copy(),
            self.seq_mappings.copy(),
            self.seq_lens.copy(),
            self.labels.copy(),
            read_ids,
            read_focus_bases,
            shuffle_on_iter=self.shuffle_on_iter,
            drop_last=self.drop_last,
            **self.clone_metadata(),
        )

    def __iter__(self):
        if self.batch_label_props is not None:
            self.shuffle_by_batch_labels()
        elif self.shuffle_on_iter:
            self.shuffle()
        self._batch_i = 0
        return self

    def __next__(self):
        if self._batch_i >= self.nbatches:
            raise StopIteration
        self._batch_i += 1
        b_st = (self._batch_i - 1) * self.batch_size
        b_en = b_st + self.batch_size
        read_ids = read_focus_bases = None
        if not self.drop_read_attrs:
            read_ids = self.read_ids[b_st:b_en]
            read_focus_bases = self.read_focus_bases[b_st:b_en]
        return (
            (
                self.sig_tensor[b_st:b_en],
                self.seq_array[b_st:b_en],
                self.seq_mappings[b_st:b_en],
                self.seq_lens[b_st:b_en],
            ),
            self.labels[b_st:b_en],
            (
                read_ids,
                read_focus_bases,
            ),
        )

    def __len__(self):
        return self.nbatches

    def get_label_counts(self):
        return Counter(self.labels[: self.nchunks])

    def split_data(
        self,
        *,
        val_prop=None,
        val_num=None,
        stratified=True,
    ):
        if val_prop is None and val_num is None:
            raise RemoraError(
                "Must supply either val_prop or val_num to split dataset"
            )
        if val_prop is not None and val_num is not None:
            LOGGER.warning(
                "val_prop and val_num supplied to split. Using val_num"
            )
            val_prop = None
        if not self.is_clipped:
            raise RemoraError("Cannot split an unclipped dataset.")
        if val_prop is not None:
            if val_prop > 0.5:
                raise RemoraError("val_prop > 0.5")
            if val_prop < 0:
                raise RemoraError("val_prop < 0")
            val_num = int(self.nchunks * val_prop)
        if val_num > self.nchunks:
            raise RemoraError("Too many validation chunks for split")
        if val_num > self.nchunks // 2:
            LOGGER.warning("Val split contains more than half of chunks")

        if not self.shuffled:
            self.shuffle()

        if stratified:
            LOGGER.debug("Performing stratified split")
            val_indices = []
            trn_indices = []
            for class_label in range(self.num_labels):
                class_indices = np.where(self.labels == class_label)[0]
                if class_indices.size == 0:
                    continue
                np.random.shuffle(class_indices)
                cls_val_idx = int(class_indices.size * val_num / self.nchunks)
                val_indices.append(class_indices[:cls_val_idx])
                trn_indices.append(class_indices[cls_val_idx:])

            val_indices = np.concatenate(val_indices)
            trn_indices = np.concatenate(trn_indices)
        else:
            val_indices = np.arange(0, val_num)
            trn_indices = np.arange(val_num, self.sig_tensor.shape[0])

        LOGGER.debug("Extracting validation dataset")
        val_read_ids = val_read_focus_bases = None
        if not self.drop_read_attrs:
            val_read_ids = self.read_ids[val_indices]
            val_read_focus_bases = self.read_focus_bases[val_indices]
        val_ds = RemoraDataset(
            self.sig_tensor[val_indices],
            self.seq_array[val_indices],
            self.seq_mappings[val_indices],
            self.seq_lens[val_indices],
            self.labels[val_indices],
            val_read_ids,
            val_read_focus_bases,
            shuffle_on_iter=False,
            drop_last=False,
            drop_read_attrs=self.drop_read_attrs,
            **self.clone_metadata(),
        )
        LOGGER.debug("Extracting train dataset")
        trn_read_ids = trn_read_focus_bases = None
        if not self.drop_read_attrs:
            trn_read_ids = self.read_ids[trn_indices]
            trn_read_focus_bases = self.read_focus_bases[trn_indices]
        trn_ds = RemoraDataset(
            self.sig_tensor[trn_indices],
            self.seq_array[trn_indices],
            self.seq_mappings[trn_indices],
            self.seq_lens[trn_indices],
            self.labels[trn_indices],
            trn_read_ids,
            trn_read_focus_bases,
            shuffle_on_iter=True,
            drop_last=False,
            drop_read_attrs=self.drop_read_attrs,
            batch_label_props=self.batch_label_props,
            **self.clone_metadata(),
        )
        return trn_ds, val_ds

    def split_by_label(self):
        if self.base_pred:
            labels = "ACGT"
        else:
            labels = ["control"] + self.mod_long_names
        label_datasets = []
        for int_label, label in enumerate(labels):
            label_indices = np.equal(self.labels, int_label)
            label_datasets.append(
                (
                    label,
                    RemoraDataset(
                        self.sig_tensor[label_indices],
                        self.seq_array[label_indices],
                        self.seq_mappings[label_indices],
                        self.seq_lens[label_indices],
                        self.labels[label_indices],
                        None
                        if self.drop_read_attrs
                        else self.read_ids[label_indices],
                        None
                        if self.drop_read_attrs
                        else self.read_focus_bases[label_indices],
                        shuffle_on_iter=False,
                        drop_last=False,
                        **self.clone_metadata(),
                    ),
                )
            )
        return label_datasets

    def balance_classes(self):
        min_class_len = min(self.get_label_counts().values())
        if self.base_pred:
            outs = 4
        else:
            outs = self.num_labels

        choices = []
        for i in range(outs):
            choices.append(
                np.random.choice(
                    np.where(self.labels == i)[0],
                    min_class_len,
                    replace=False,
                )
            )
        choices = np.concatenate(choices)

        return RemoraDataset(
            sig_tensor=self.sig_tensor[choices],
            seq_array=self.seq_array[choices],
            seq_mappings=self.seq_mappings[choices],
            seq_lens=self.seq_lens[choices],
            labels=self.labels[choices],
            read_ids=None if self.drop_read_attrs else self.read_ids[choices],
            read_focus_bases=None
            if self.drop_read_attrs
            else self.read_focus_bases[choices],
            nchunks=outs * min_class_len,
            **self.clone_metadata(),
        )

    def filter(self, indices):
        if len(indices) > self.sig_tensor.shape[0]:
            raise RemoraError(
                "Filter indices cannot be longer than dataset size"
            )
        return RemoraDataset(
            self.sig_tensor[indices],
            self.seq_array[indices],
            self.seq_mappings[indices],
            self.seq_lens[indices],
            self.labels[indices],
            None if self.drop_read_attrs else self.read_ids[indices],
            None if self.drop_read_attrs else self.read_focus_bases[indices],
            shuffle_on_iter=False,
            drop_last=False,
            **self.clone_metadata(),
        )

    def add_fake_base(self, new_mod_long_names, new_mod_bases):
        if not set(self.mod_long_names).issubset(new_mod_long_names):
            raise RemoraError(
                "There are no mods in common between the model being trained "
                f"({new_mod_long_names}) and the external validation set "
                f"({self.mod_long_names})."
            )
        for mod in self.mod_long_names:
            new_index = new_mod_long_names.index(mod) + 1
            old_index = self.mod_long_names.index(mod) + 1
            self.labels[np.where(self.labels == old_index)] = new_index
        self.mod_long_names = new_mod_long_names
        self.mod_bases = new_mod_bases

    def save(self, filename):
        self.clip_chunks()
        np.savez(
            filename,
            sig_tensor=self.sig_tensor,
            seq_array=self.seq_array,
            seq_mappings=self.seq_mappings,
            seq_lens=self.seq_lens,
            labels=self.labels,
            read_ids=None if self.drop_read_attrs else self.read_ids,
            read_focus_bases=None
            if self.drop_read_attrs
            else self.read_focus_bases,
            chunk_context=self.chunk_context,
            kmer_context_bases=self.kmer_context_bases,
            base_pred=self.base_pred,
            mod_bases=self.mod_bases,
            mod_long_names=self.mod_long_names,
            motifs=[mot[0] for mot in self.motifs],
            motif_offset=[mot[1] for mot in self.motifs],
            reverse_signal=int(self.reverse_signal),
            base_start_justify=int(self.base_start_justify),
            offset=self.offset,
            version=DATASET_VERSION,
            **self.sig_map_refiner.get_save_kwargs(),
        )

    @classmethod
    def load_from_file(cls, filename, *args, **kwargs):
        # use allow_pickle=True to allow None type in read_data
        data = np.load(filename, allow_pickle=True)
        try:
            version = int(data["version"].item())
        except KeyError:
            version = None
        if version is None or version != DATASET_VERSION:
            raise RemoraError(
                f"Remora dataset version ({version}) does "
                f"not match current distribution ({DATASET_VERSION})"
            )
        chunk_context = tuple(data["chunk_context"].tolist())
        kmer_context_bases = tuple(data["kmer_context_bases"].tolist())
        base_pred = data["base_pred"].item()
        try:
            mod_bases = data["mod_bases"].item()
        except ValueError:
            mod_bases = ""
        mod_long_names = tuple(data["mod_long_names"].tolist())
        if isinstance(data["motif_offset"].tolist(), int):
            motifs = [
                (mot, mot_off)
                for mot, mot_off in zip(
                    [data["motif"].tolist()], [data["motif_offset"].tolist()]
                )
            ]
        else:
            motifs = [
                (mot, mot_off)
                for mot, mot_off in zip(
                    data["motifs"].tolist(), data["motif_offset"].tolist()
                )
            ]
        sig_map_refiner = SigMapRefiner.load_from_np_savez(data)
        base_start_justify = bool(int(data["base_start_justify"].item()))
        # default to reverse_signal=False for datasets without this attribute
        reverse_signal = bool(int(data.get("reverse_signal", 0)))
        offset = int(data["offset"].item())
        return cls(
            data["sig_tensor"],
            data["seq_array"],
            data["seq_mappings"],
            data["seq_lens"],
            data["labels"],
            data["read_ids"],
            data["read_focus_bases"],
            chunk_context=chunk_context,
            kmer_context_bases=kmer_context_bases,
            base_pred=base_pred,
            mod_bases=mod_bases,
            mod_long_names=mod_long_names,
            motifs=motifs,
            reverse_signal=reverse_signal,
            sig_map_refiner=sig_map_refiner,
            base_start_justify=base_start_justify,
            offset=offset,
            *args,
            **kwargs,
        )

    def perturb_seq_mismatch(self, mm_rate):
        errs_added = 0
        for c_idx, c_num_mm in tqdm(
            enumerate(np.random.binomial(self.seq_lens, mm_rate)),
            smoothing=0,
            total=self.nchunks,
            desc="Mismatches",
            leave=False,
            disable=os.environ.get("LOG_SAFE", False),
        ):
            if c_num_mm == 0:
                continue
            for seq_pos in np.random.choice(
                self.seq_lens[c_idx], c_num_mm, replace=False
            ):
                # convert from position in chunk sequence to position in
                # sequence array (including k-mer context bases).
                seq_arr_pos = seq_pos + self.kmer_context_bases[0]
                if self.seq_array[c_idx, seq_arr_pos] == -1:
                    continue
                self.seq_array[c_idx, seq_arr_pos] = np.random.choice(
                    MISMATCH_ARRS[self.seq_array[c_idx, seq_arr_pos]]
                )
                errs_added += 1
        LOGGER.info(f"Introduced {errs_added} mismatch errors")

    def perturb_seq_to_sig_map(self, sig_shift):
        for c_idx, c_seq_len in tqdm(
            enumerate(self.seq_lens),
            smoothing=0,
            total=self.nchunks,
            desc="Signal shifts",
            leave=False,
            disable=os.environ.get("LOG_SAFE", False),
        ):
            # shift seq to sig mapping in the middle of chunk
            self.seq_mappings[c_idx, 1:c_seq_len] = np.clip(
                self.seq_mappings[c_idx, 1:c_seq_len] + sig_shift,
                self.seq_mappings[c_idx, 0],
                self.seq_mappings[c_idx, c_seq_len],
            )

    @property
    def can_base(self):
        return self.motifs[0][0][self.motifs[0][1]]

    @property
    def is_clipped(self):
        return self.nchunks == self.sig_tensor.shape[0]

    @property
    def is_multiclass(self):
        return self.base_pred or len(self.mod_bases) > 1

    @property
    def num_motifs(self):
        return len(self.motifs)

    @property
    def num_labels(self):
        if self.base_pred:
            return 4
        return len(self.mod_bases) + 1

    @property
    def summary(self):
        return (
            f"               num chunks : {self.nchunks}\n"
            f"       label distribution : {self.get_label_counts()}\n"
            f"                base_pred : {self.base_pred}\n"
            f"                mod_bases : {self.mod_bases}\n"
            f"           mod_long_names : {self.mod_long_names}\n"
            f"       kmer_context_bases : {self.kmer_context_bases}\n"
            f"            chunk_context : {self.chunk_context}\n"
            f"                   motifs : {self.motifs}\n"
            f"           reverse_signal : {self.reverse_signal}\n"
            f" chunk_extract_base_start : {self.base_start_justify}\n"
            f"     chunk_extract_offset : {self.offset}\n"
            f"          sig_map_refiner : {self.sig_map_refiner}\n"
        )

    def __repr__(self):
        return self.summary

    @classmethod
    def allocate_empty_chunks(
        cls,
        num_chunks,
        chunk_context,
        kmer_context_bases,
        max_seq_len=None,
        min_samps_per_base=None,
        *args,
        **kwargs,
    ):
        if max_seq_len is None and min_samps_per_base is None:
            raise RemoraError(
                "Must set either max_seq_len or min_samps_per_base"
            )
        if max_seq_len is None:
            max_seq_len = sum(chunk_context) // min_samps_per_base
        sig_tensor = np.empty(
            (num_chunks, 1, sum(chunk_context)), dtype=np.float32
        )
        seq_array = np.empty(
            (num_chunks, max_seq_len + sum(kmer_context_bases)),
            dtype=np.byte,
        )
        seq_mappings = np.empty(
            (num_chunks, max_seq_len + 1),
            dtype=np.short,
        )
        seq_lens = np.empty(num_chunks, dtype=np.short)
        labels = np.empty(num_chunks, dtype=np.int64)
        read_ids = np.empty(num_chunks, dtype="U36")
        read_pos = np.empty(num_chunks, dtype=int)
        return cls(
            sig_tensor,
            seq_array,
            seq_mappings,
            seq_lens,
            labels,
            read_ids,
            read_pos,
            nchunks=0,
            chunk_context=chunk_context,
            max_seq_len=max_seq_len,
            kmer_context_bases=kmer_context_bases,
            *args,
            **kwargs,
        )


def merge_datasets(input_datasets, balance=False, quiet=False):
    def load_dataset(ds_path, num_chunks):
        dataset = RemoraDataset.load_from_file(
            ds_path,
            shuffle_on_iter=False,
            drop_last=False,
        )
        if num_chunks is not None and num_chunks < dataset.nchunks:
            dataset.shuffle()
        else:
            num_chunks = dataset.nchunks
        return dataset, num_chunks

    log_fp = LOGGER.debug if quiet else LOGGER.info

    # load first file to determine base_pred or mod_bases
    dataset, num_chunks = load_dataset(*input_datasets[0])
    base_pred = dataset.base_pred
    chunk_context = dataset.chunk_context
    max_seq_len = dataset.max_seq_len
    kmer_context_bases = dataset.kmer_context_bases
    motifs = set(dataset.motifs)
    sig_map_refiner = dataset.sig_map_refiner
    base_start_justify = dataset.base_start_justify
    reverse_signal = dataset.reverse_signal
    offset = dataset.offset
    raw_mod_long_names = []
    raw_mod_bases = []
    if not base_pred:
        for mod_base, mln in zip(dataset.mod_bases, dataset.mod_long_names):
            if mod_base not in raw_mod_bases:
                raw_mod_bases.append(mod_base)
                raw_mod_long_names.append(mln)
    del dataset
    gc.collect()

    all_num_chunks = [num_chunks]
    for ds_path, num_chunks in input_datasets[1:]:
        dataset, num_chunks = load_dataset(ds_path, num_chunks)
        for attr_name, attr_val in (
            ("base_pred", base_pred),
            ("chunk_context", chunk_context),
            ("max_seq_len", max_seq_len),
            ("kmer_context_bases", kmer_context_bases),
            ("base_start_justify", base_start_justify),
            ("offset", offset),
        ):
            if getattr(dataset, attr_name) != attr_val:
                raise RemoraError(
                    f"All datasets must have same {attr_name} "
                    f"{getattr(dataset, attr_name)} != {attr_val}"
                )
        if set(dataset.motifs) != motifs:
            log_fp(
                "WARNING: Datasets have different motifs. Merging motifs "
                f"{motifs} with motifs {set(dataset.motifs)}"
            )
            motifs.update(dataset.motifs)
        if not base_pred:
            for mod_base, mln in zip(dataset.mod_bases, dataset.mod_long_names):
                if mod_base not in raw_mod_bases:
                    raw_mod_bases.append(mod_base)
                    raw_mod_long_names.append(mln)
        all_num_chunks.append(num_chunks)
        # TODO add checks for refine attributes
        del dataset
        gc.collect()

    all_mod_bases = ""
    all_mod_long_names = None
    if not base_pred:
        # sort modified bases alphabetically
        all_mod_long_names = []
        for idx in sorted(
            range(len(raw_mod_bases)), key=raw_mod_bases.__getitem__
        ):
            all_mod_bases += raw_mod_bases[idx]
            all_mod_long_names.append(raw_mod_long_names[idx])

    output_dataset = RemoraDataset.allocate_empty_chunks(
        num_chunks=sum(all_num_chunks),
        chunk_context=chunk_context,
        max_seq_len=max_seq_len,
        kmer_context_bases=kmer_context_bases,
        base_pred=base_pred,
        mod_bases=all_mod_bases,
        mod_long_names=all_mod_long_names,
        motifs=motifs,
        reverse_signal=reverse_signal,
        sig_map_refiner=sig_map_refiner,
        base_start_justify=base_start_justify,
        offset=offset,
    )
    LOGGER.info(f"Output dataset summary:\n{output_dataset.summary}")
    for ds_path, num_chunks in input_datasets:
        input_dataset, num_chunks = load_dataset(ds_path, num_chunks)
        if base_pred:
            label_conv = np.arange(4, dtype=np.int64)
        else:
            label_conv = np.empty(
                len(input_dataset.mod_bases) + 1, dtype=np.int64
            )
            label_conv[0] = 0
            for input_lab, mod_base in enumerate(input_dataset.mod_bases):
                label_conv[input_lab + 1] = (
                    output_dataset.mod_bases.find(mod_base) + 1
                )
        added_chunks = 0
        for (
            (b_sig, b_seq, b_ss_map, b_seq_lens),
            b_labels,
            (b_rids, b_rfbs),
        ) in input_dataset:
            b_labels = label_conv[b_labels]
            if added_chunks + b_labels.size >= num_chunks:
                batch_size = num_chunks - added_chunks
                output_dataset.add_batch(
                    b_sig[:batch_size],
                    b_seq[:batch_size],
                    b_ss_map[:batch_size],
                    b_seq_lens[:batch_size],
                    b_labels[:batch_size],
                    b_rids[:batch_size],
                    b_rfbs[:batch_size],
                )
                added_chunks += batch_size
                break
            added_chunks += b_labels.size
            output_dataset.add_batch(
                b_sig, b_seq, b_ss_map, b_seq_lens, b_labels, b_rids, b_rfbs
            )
        log_fp(
            f"Copied {added_chunks} chunks. New label distribution: "
            f"{output_dataset.get_label_counts()}"
        )
        del input_dataset
        gc.collect()
    output_dataset.clip_chunks()
    if balance:
        balanced_dataset = output_dataset.balance_classes()
        log_fp(
            f"Balanced out to {balanced_dataset.sig_tensor.shape[0]} chunks. "
            f"New label distribution: {balanced_dataset.get_label_counts()}"
        )
        return balanced_dataset

    return output_dataset



from itertools import product
from dataclasses import dataclass, field

import numpy as np
from scipy import stats

from remora import RemoraError, log
from remora.constants import (
    DEFAULT_REFINE_HBW,
    DEFAULT_REFINE_SHORT_DWELL_PARAMS,
    DEFAULT_REFINE_ALGO,
    REFINE_ALGO_DWELL_PEN_NAME,
)
from remora.refine_signal_map_core import (
    seq_banded_dp,
    extract_levels,
    adjust_seq_band,
)

LOGGER = log.get_logger()


###############
# Short dwell #
###############


def compute_dwell_pen_array(target, limit, weight):
    if limit > target:
        LOGGER.warning(
            f"Requested short dwell limit ({limit}) is greater than target "
            f"dwell ({target}). Setting limit to target."
        )
        limit = target
    return weight * np.square(np.arange(limit, dtype=np.float32) - target)


DEFAULT_REFINE_SHORT_DWELL_PEN = compute_dwell_pen_array(
    *DEFAULT_REFINE_SHORT_DWELL_PARAMS
)


#################
# Re-scale Core #
#################


def rescale_lstsq(dacs, levels, shift, scale):
    norm_sig = (dacs - shift) / scale
    shift_est, scale_est = np.linalg.lstsq(
        np.column_stack([np.ones_like(norm_sig), norm_sig]),
        levels,
        rcond=None,
    )[0]
    if scale_est == 0:
        return shift, scale
    new_shift = shift - (scale * shift_est / scale_est)
    new_scale = scale / scale_est
    return new_shift, new_scale


def rough_rescale_lstsq(dacs, levels, shift, scale, quants):
    norm_sig = (dacs - shift) / scale
    norm_qs = np.quantile(norm_sig, quants)
    shift_est, scale_est = np.linalg.lstsq(
        np.column_stack([np.ones_like(norm_qs), norm_qs]),
        np.quantile(levels, quants),
        rcond=None,
    )[0]
    if scale_est == 0:
        return shift, scale
    new_shift = shift - (scale * shift_est / scale_est)
    new_scale = scale / scale_est
    return new_shift, new_scale


##########################
# Signal Mapping Refiner #
##########################


def index_from_kmer(kmer, alphabet="ACGT"):
    """Encode string k-mer as integer via len(alphabet)-bit encoding.

    Args:
        kmer (str): kmer string
        alphabet (str): bases used. Default: ACGT

    Returns:
        int: bit encoded kmer

    Example:
        index_from_kmer('AAA', 'ACG')               returns 0
        index_from_kmer('CAAAAAAAA', 'ACGTVWXYX')   returns 65536
    """
    return sum(
        alphabet.find(base) * (len(alphabet) ** kmer_pos)
        for kmer_pos, base in enumerate(kmer[::-1])
    )


@dataclass
class SigMapRefiner:
    """Object to perform scaling and signal mapping refinement before chunk
    extraction.

    Args:
        TODO
    """

    kmer_model_filename: str = None
    do_rough_rescale: bool = True
    scale_iters: int = -1
    algo: str = DEFAULT_REFINE_ALGO
    half_bandwidth: int = DEFAULT_REFINE_HBW
    sd_params: tuple = None
    do_fix_guage: bool = False

    sd_arr: np.ndarray = field(
        default_factory=lambda: DEFAULT_REFINE_SHORT_DWELL_PEN
    )
    _levels_array: np.ndarray = None
    str_kmer_levels: dict = None
    kmer_len: int = None
    center_idx: int = -1
    is_loaded: bool = False

    def __repr__(self):
        if not self.is_loaded:
            return "No Remora signal refine/map settings loaded"
        r_str = (
            f"Loaded {self.kmer_len}-mer table with {self.center_idx + 1} "
            "central position."
        )
        if self.do_rough_rescale:
            r_str += " Rough re-scaling will be executed."
        if self.scale_iters > 0:
            r_str += (
                f" {self.scale_iters} rounds of signal mapping refinement "
                "followed by precise re-scaling will be executed."
            )
        if self.scale_iters >= 0:
            r_str += (
                " Signal mapping refinement will be executed using the "
                f"{self.algo} refinement method (band half width: "
                f"{self.half_bandwidth})."
            )
            if self.algo == REFINE_ALGO_DWELL_PEN_NAME:
                r_str += f" Short dwell penalty array set to {self.sd_arr}."
        return r_str

    @property
    def bases_before(self):
        """Number of bases in k-mer before the central base"""
        return self.center_idx

    @property
    def bases_after(self):
        """Number of bases in k-mer after the central base"""
        return self.kmer_len - self.center_idx - 1

    def write_kmer_table(self, fh):
        for kmer in product(*["ACGT"] * self.kmer_len):
            fh.write(
                f"{''.join(kmer)}\t"
                f"{self._levels_array[index_from_kmer(kmer)]}\n"
            )

    def load_kmer_table(self):
        self.str_kmer_levels = {}
        with open(self.kmer_model_filename) as kmer_fp:
            self.kmer_len = len(kmer_fp.readline().split()[0])
            kmer_fp.seek(0)
            for line in kmer_fp:
                kmer, level = line.split()
                kmer = kmer.upper()
                if kmer in self.str_kmer_levels:
                    raise RemoraError(
                        f"K-mer found twice in levels file '{kmer}'."
                    )
                if self.kmer_len != len(kmer):
                    raise RemoraError(
                        f"K-mer lengths not all equal '{len(kmer)} != "
                        f"{self.kmer_len}' for {kmer}."
                    )
                try:
                    self.str_kmer_levels[kmer] = float(level)
                    if np.isnan(self.str_kmer_levels[kmer]):
                        self.str_kmer_levels[kmer] = 0
                except ValueError:
                    raise RemoraError(
                        f"Could not convert level to float '{level}'"
                    )
        if len(self.str_kmer_levels) != 4**self.kmer_len:
            raise RemoraError(
                "K-mer table contains fewer entries "
                f"({len(self.str_kmer_levels)}) than expected "
                f"({4 ** self.kmer_len})"
            )

    def determine_dominant_pos(self):
        if self.str_kmer_levels is None:
            return
        sorted_kmers = sorted(
            (level, kmer) for kmer, level in self.str_kmer_levels.items()
        )
        kmer_idx_stats = []
        kmer_summ = ""
        for kmer_idx in range(self.kmer_len):
            kmer_idx_pos = []
            for base in "ACGT":
                kmer_idx_pos.append(
                    [
                        levels_idx
                        for levels_idx, (_, kmer) in enumerate(sorted_kmers)
                        if kmer[kmer_idx] == base
                    ]
                )
            # compute Kruskal-Wallis H-test statistics for non-random ordering
            # of groups, indicating the dominant position within the k-mer
            kmer_idx_stats.append(stats.kruskal(*kmer_idx_pos)[0])
            kmer_summ += f"\t{kmer_idx}\t{kmer_idx_stats[-1]:10.2f}\n"
        self.center_idx = np.argmax(kmer_idx_stats)
        LOGGER.debug(f"K-mer index stats:\n{kmer_summ}")
        LOGGER.debug(f"Choosen central position: {self.center_idx}")

    def __post_init__(self):
        # determine if level model is loaded from any of 3 options
        if self._levels_array is not None and not np.array_equal(
            self._levels_array, np.array(None)
        ):
            self.is_loaded = True
            self.kmer_len = int(np.log(self._levels_array.size) / np.log(4))
            assert 4**self.kmer_len == self._levels_array.size
        elif self.kmer_model_filename is not None:
            self.load_kmer_table()
            self.is_loaded = True
            self.determine_dominant_pos()
            if self.do_fix_guage:
                self.fix_gauge()
        elif self.str_kmer_levels is not None:
            self.is_loaded = True
            self.determine_dominant_pos()
            if self.do_fix_guage:
                self.fix_gauge()
        if not self.is_loaded and (
            self.do_rough_rescale or self.scale_iters >= 0
        ):
            raise RemoraError(
                "Signal re-scaling is requested without levels table."
            )

        if self.sd_params is not None:
            self.sd_arr = compute_dwell_pen_array(*self.sd_params)
        if (
            self.is_loaded
            and self.scale_iters >= 0
            and self.algo == REFINE_ALGO_DWELL_PEN_NAME
        ):
            LOGGER.debug(f"Refine short dwell penalty array: {self.sd_arr}")
        if (
            self.is_loaded
            and not self.do_rough_rescale
            and self.scale_iters < 0
        ):
            LOGGER.warning(
                "K-mer table provided, but not used. "
                "See rough rescaling options."
            )

    def extract_levels(self, int_seq):
        return extract_levels(
            int_seq.astype(np.int32),
            self.levels_array,
            self.kmer_len,
            self.center_idx,
        )

    def fix_gauge(self):
        med = np.median(self.levels_array)
        # note factor to scale MAD to SD
        mad = np.median(np.absolute(self.levels_array - med)) * 1.4826
        self._levels_array = (self.levels_array - med) / mad
        self.str_kmer_levels = {}
        for kmer in product(*["ACGT"] * self.kmer_len):
            self.str_kmer_levels["".join(kmer)] = self._levels_array[
                index_from_kmer(kmer)
            ]

    @property
    def levels_array(self):
        if self._levels_array is None:
            if self.str_kmer_levels is None:
                return
            self._levels_array = np.empty(4**self.kmer_len, dtype=np.float32)
            for kmer, level in self.str_kmer_levels.items():
                self._levels_array[index_from_kmer(kmer)] = level
        return self._levels_array

    def rough_rescale(
        self,
        shift,
        scale,
        seq_to_sig_map,
        int_seq,
        dacs,
        quants=np.arange(0.05, 1, 0.05),
        clip_bases=10,
        use_base_center=True,
    ):
        """Estimate new scaling parameters base on quantiles of levels and
        quantiles of central signal point in each base.

        Args:
            levels (np.array): Estimated reference levels for each base
            quants (np.array): Quantiles to use for rough estimates
            clip_bases (int): Number of bases to clip from either end before
                computing rough rescale estimates.
            use_base_center (bool): Use the central signal point from each base.
                If False, quantiles will be computed from the full signal.
        """
        levels = self.extract_levels(int_seq)
        if use_base_center:
            optim_dacs = dacs[(seq_to_sig_map[:-1] + seq_to_sig_map[1:]) // 2]
            if clip_bases > 0 and levels.size > clip_bases * 2:
                levels = levels[clip_bases:-clip_bases]
                optim_dacs = optim_dacs[clip_bases:-clip_bases]
        else:
            optim_dacs = dacs[seq_to_sig_map[0] : seq_to_sig_map[-1]]
        return rough_rescale_lstsq(optim_dacs, levels, shift, scale, quants)

    def rescale(
        self,
        levels,
        dacs,
        shift,
        scale,
        seq_to_sig_map,
        dwell_filter_pctls=(10, 90),
        min_abs_level=0.2,
        edge_filter_bases=10,
        min_levels=10,
    ):
        """Estimate new scaling parameters base on current signal mapping to
        estimated levels.

        Args:
            levels (np.array): Estimated reference levels for each base
            dacs (np.ndarray): Unnormalized DAC signal
            shift (float): Shift from dac to normalized signal. via formula:
                norm = (dac - shift) / scale
            scale (float): Scale from dac to normalized signal
            seq_to_sig_map (np.ndarray): Position within signal array assigned
                to each base in seq
            dwell_filter_pctls (tuple): Lower and upper percentile values to
                filter short and stalled bases from estimation
            min_abs_level (float): Minimum (absolute values) level to include
                in computaiton
            edge_filter_bases (int): Number of bases at the edge of a read to
                remove from optimization computation
            min_levels (int): Minimum number of levels to perform re-scaling
        """
        with np.errstate(invalid="ignore"):
            dacs_cumsum = np.empty(dacs.size + 1)
            dacs_cumsum[0] = 0
            dacs_cumsum[1:] = np.cumsum(dacs)
            dwells = np.diff(seq_to_sig_map)
            dac_means = np.diff(dacs_cumsum[seq_to_sig_map]) / dwells

        # filter bases base on lower and upper percentile of dwell to remove
        # regions of poor signal assignment
        # estimate dwell limits (these limits are treated as exclusive
        # boundaries for filtering as the lower boundary may be 1)
        dwells = np.diff(seq_to_sig_map)
        dwell_min, dwell_max = np.percentile(dwells, dwell_filter_pctls)
        edge_filter = np.full(dwells.size, True, dtype=np.bool)
        if edge_filter_bases > 0:
            edge_filter[:edge_filter_bases] = False
            edge_filter[-edge_filter_bases:] = False
        valid_bases = np.logical_and.reduce(
            (
                dwells > dwell_min,
                dwells < dwell_max,
                np.abs(levels - np.mean(levels)) > min_abs_level,
                np.logical_not(np.isnan(dac_means)),
                edge_filter,
            )
        )
        filt_levels = levels[valid_bases]
        filt_dacs = dac_means[valid_bases]
        if filt_levels.size < min_levels:
            raise RemoraError("Too few positions")

        return rescale_lstsq(filt_dacs, filt_levels, shift, scale)

    def refine_sig_map(self, shift, scale, seq_to_sig_map, int_seq, dacs):
        levels = self.extract_levels(int_seq)
        dacs = dacs[seq_to_sig_map[0] : seq_to_sig_map[-1]]
        sig_st = seq_to_sig_map[0]
        seq_to_sig_map = seq_to_sig_map - sig_st
        # 0 indicates 1 round of sig map refine w/o rescaling
        for _ in range(max(1, self.scale_iters)):
            seq_to_sig_map, _, _, _, _ = refine_signal_mapping(
                (dacs - shift) / scale,
                seq_to_sig_map,
                levels,
                self.half_bandwidth,
                self.algo,
                self.sd_arr,
            )
            # 0 specifies no re-scaling
            if self.scale_iters > 0:
                try:
                    shift, scale = self.rescale(
                        levels, dacs, shift, scale, seq_to_sig_map
                    )
                except RemoraError as e:
                    LOGGER.debug(f"Remora rescaling error: {self.read_id} {e}")
                    break
        return seq_to_sig_map + sig_st, shift, scale

    def get_save_kwargs(self):
        return {
            "refine_kmer_levels": self._levels_array,
            "refine_kmer_center_idx": self.center_idx,
            "refine_do_rough_rescale": int(self.do_rough_rescale),
            "refine_scale_iters": self.scale_iters,
            "refine_algo": self.algo,
            "refine_half_bandwidth": self.half_bandwidth,
            "refine_sd_arr": self.sd_arr,
        }

    @classmethod
    def load_from_np_savez(cls, data):
        return cls(
            _levels_array=data["refine_kmer_levels"],
            center_idx=int(data["refine_kmer_center_idx"].item()),
            do_rough_rescale=bool(int(data["refine_do_rough_rescale"].item())),
            scale_iters=int(data["refine_scale_iters"].item()),
            algo=str(data["refine_algo"].item()),
            half_bandwidth=int(data["refine_half_bandwidth"].item()),
            sd_arr=data["refine_sd_arr"],
        )

    @classmethod
    def load_from_dict(
        cls,
        data,
        do_rough_rescale=True,
        scale_iters=-1,
        algo=DEFAULT_REFINE_ALGO,
        half_bandwidth=DEFAULT_REFINE_HBW,
        sd_params=None,
        do_fix_guage=False,
        sd_arr=DEFAULT_REFINE_SHORT_DWELL_PEN,
    ):
        """Create refiner from str_kmer_levels dict with kmer keys and float
        current level values.
        """
        kmer_len = len(next(iter(data.keys())))
        return cls(
            do_rough_rescale=do_rough_rescale,
            scale_iters=scale_iters,
            algo=algo,
            half_bandwidth=half_bandwidth,
            sd_params=sd_params,
            do_fix_guage=do_fix_guage,
            sd_arr=sd_arr,
            str_kmer_levels=data,
            kmer_len=kmer_len,
        )


###########
# Banding #
###########


def compute_sig_band(bps, levels, bhw=DEFAULT_REFINE_HBW, is_banded=True):
    """Compute band over which to explore possible paths. Band is represented
    in sequence/level coordinates at each signal position.

    Args:
        bps (np.ndarray): Integer array containing breakpoints
        levels (np.ndarray): float array containing expected signal levels. May
            contain np.NAN values. Band will be constructed to maintain path
            through NAN regions.
        bhw (int): band half width
        is_banded (bool): Should bhw be applied over full path. If False only
            NAN levels restrict the computed band.

    Returns:
        int32 np.ndarray with shape (2, sig_len = bps[-1] - bps[0]). The first
        row contains the lower band boundaries in sequence coordinates and the
        second row contains the upper boundaries in sequence coordinates.
    """
    if is_banded and bhw is None:
        raise RemoraError("Cannot compute band with half width of None.")
    seq_len = levels.size
    if bps.size - 1 != seq_len:
        raise RemoraError("Breakpoints must be one longer than levels.")
    sig_len = bps[-1] - bps[0]
    seq_indices = np.repeat(np.arange(seq_len), np.diff(bps))

    # Calculate bands
    # The 1st row consists of the start indices (inc) and the 2nd row
    # consists of the end indices (exc) of the valid rows for each col.
    band = np.empty((2, sig_len), dtype=np.int32)
    if is_banded:
        # Use specific band if instructed
        band[0, :] = np.maximum(seq_indices - bhw, 0)
        band[1, :] = np.minimum(seq_indices + bhw + 1, seq_len)
    else:
        # otherwise specify entire input matrix
        band[0, :] = 0
        band[1, :] = seq_len

    # Modify bands based on invalid levels
    nan_mask = np.in1d(seq_indices, np.nonzero(np.isnan(levels)))
    nan_sig_indices = np.where(nan_mask)[0]
    nan_seq_indices = seq_indices[nan_mask]
    band[0, nan_sig_indices] = nan_seq_indices
    band[1, nan_sig_indices] = nan_seq_indices + 1
    # Modify bands close to invalid levels so monotonically increasing
    band[0, :] = np.maximum.accumulate(band[0, :])
    band[1, :] = np.minimum.accumulate(band[1, ::-1])[::-1]

    return band


def validate_band(band, sig_len=None, seq_len=None, is_sig_band=True):
    """Validate that band is valid and agrees with input data.

    Args:
        band (np.array): int32 array with shape (2, sig_len or seq_len). The
            first row contains the lower band boundaries and the second row
            contains the upper boundaries.
        sig_len (int): Length of signal associated with band
        seq_len (int): Length of sequence/levels associated with band
        is_sig_band (bool): Does the provided band specify sequence/level
            positions for each signal position? If not it is assumed that the
            band contains signal positions for each sequence/level position.

    Raises:
        RemoraError if any portion of the band is determined to be invalid.
    """
    # first coordinate 0, last coordinate signal length
    if band[0, 0] != 0:
        raise RemoraError("Band does not start with 0 coordinate.")

    # ends all greater than starts
    if np.diff(band, axis=0)[0].min() <= 0:
        raise RemoraError("Band contains 0-length region")
    # monotonic start and end postions
    if np.diff(band[0]).min() < 0:
        raise RemoraError(
            "Band start positions are not monotonically increasing"
        )
    if np.diff(band[1]).min() < 0:
        raise RemoraError("Band end positions are not monotonically increasing")

    # if provided check that start and end coordinates agree with signal and
    # levels.
    if is_sig_band:
        if sig_len is not None and band.shape[1] != sig_len:
            LOGGER.debug(f"Invalid sig_band length: {band.shape[1]} {sig_len}")
            raise RemoraError("Invalid sig_band length")
        if seq_len is not None and band[1, -1] != seq_len:
            LOGGER.debug(
                f"Invalid sig_band end coordinate: {band[1, -1]} {seq_len}"
            )
            raise RemoraError("Invalid sig_band end coordinate")
    else:
        if sig_len is not None and band[1, -1] != sig_len:
            LOGGER.debug(
                f"Invalid seq_band end coordinate: {band[1, -1]} {sig_len}"
            )
            raise RemoraError("Invalid seq_band end coordinate")
        if seq_len is not None and band.shape[1] != seq_len:
            LOGGER.debug(f"Invalid sig_band length: {band.shape[1]} {seq_len}")
            raise RemoraError("Invalid sig_band length")


def convert_to_seq_band(sig_band):
    """Convert band with sig_len entries containing upper and lower band
    boundaries in base coordinates to a seq_len entries contraining upper and
    lower band boundaries in signal space.

    Args:
        sig_band (np.array): int32 array with shape (2, sig_len). The first row
            contains the lower band boundaries in sequence coordinates and the
            second row contains the upper boundaries in sequence coordinates.

    Returns:
        int32 np.ndarray with shape (2, seq_len = sig_band[1, -1]). The first
        row contains the lower band boundaries in signal coordinates and the
        second row contains the upper boundaries in signal coordinates.
    """
    sig_len = sig_band.shape[1]
    seq_len = sig_band[1, -1]
    seq_band = np.zeros((2, seq_len), dtype=np.int32)
    seq_band[1, :] = sig_len

    # upper signal coordinates define lower sequence boundaries
    lower_sig_pos = np.nonzero(np.ediff1d(sig_band[1, :], to_begin=0))[0]
    lower_base_pos = sig_band[1, lower_sig_pos - 1]
    seq_band[0, lower_base_pos] = lower_sig_pos
    seq_band[0, :] = np.maximum.accumulate(seq_band[0, :])

    upper_sig_pos = np.nonzero(np.ediff1d(sig_band[0, :], to_begin=0))[0]
    upper_base_pos = sig_band[0, upper_sig_pos]
    seq_band[1, upper_base_pos - 1] = upper_sig_pos
    seq_band[1, :] = np.minimum.accumulate(seq_band[1, ::-1])[::-1]

    return seq_band


######################
# Signal Map Refiner #
######################


def refine_signal_mapping(
    signal,
    seq_to_sig_map,
    levels,
    band_half_width=DEFAULT_REFINE_HBW,
    refine_algo=DEFAULT_REFINE_ALGO,
    short_dwell_pen=DEFAULT_REFINE_SHORT_DWELL_PEN,
    adjust_band_min_step=2,
):
    """Refine input signal mapping to minimize difference difference between
    signal and levels.

    Args:
        signal (np.ndarray): Float32 array containing normalized signal values.
        seq_to_sig_map (np.ndarray): Int32 array with locations within
            signal for each base represented by levels. `seq_to_sig_map.size`
            should equal `levels.size + 1` in order to include the end of the
            last base. Values should be monotonically increasing (ties allowed).
        levels (np.ndarray): Float32 array containing estimated levels.
        band_half_width (int): Half bandwidth
        adjust_band_min_step (int): Minimum step between one base and the next
            to enforce in band adjustment.

    Returns:
        2-tuple containing:
            1. Signal mapping
            2. Mapping score
    """
    # Note that there is no guarantee the seq_to_sig_map go right to the end
    # of the signal provided so may need to trim signal
    signal = signal[seq_to_sig_map[0] : seq_to_sig_map[-1]]
    sig_map_start = 0
    if seq_to_sig_map[0] != 0:
        sig_map_start = seq_to_sig_map[0]
        seq_to_sig_map = np.copy(seq_to_sig_map)
        seq_to_sig_map -= seq_to_sig_map[0]

    # TODO compute seq band directly with anti-diagonal band
    sig_band = compute_sig_band(
        seq_to_sig_map,
        levels,
        bhw=band_half_width,
    )
    seq_band = convert_to_seq_band(sig_band)
    adjust_seq_band(seq_band, min_step=adjust_band_min_step)
    validate_band(
        seq_band,
        sig_len=signal.shape[0],
        seq_len=levels.shape[0],
        is_sig_band=False,
    )

    # Change nans to zeros so nans don't break things
    temp_levels = np.copy(levels)
    temp_levels[np.isnan(levels)] = 0
    all_scores, path, traceback, base_offsets = seq_banded_dp(
        signal.astype(np.float32),
        temp_levels.astype(np.float32),
        seq_band,
        short_dwell_pen,
        refine_algo,
    )
    return path + sig_map_start, all_scores, traceback, seq_band, base_offsets



import os
import re
import random
from pathlib import Path
from typing import Callable
from copy import copy, deepcopy
from dataclasses import dataclass
from collections import defaultdict
from itertools import chain, product
from functools import cached_property

import pod5
import pysam
import numpy as np
import pandas as pd
from tqdm import tqdm
import seaborn as sns
from pysam import AlignedSegment
from matplotlib import pyplot as plt

from remora.metrics import METRIC_FUNCS
from remora.constants import PA_TO_NORM_SCALING_FACTOR, DEFAULT_PLOT_FIG_SIZE
from remora import log, util, data_chunks as DC, duplex_utils as DU, RemoraError

LOGGER = log.get_logger()

_SIG_PROF_FN = os.getenv("REMORA_EXTRACT_SIGNAL_PROFILE_FILE")
_ALIGN_PROF_FN = os.getenv("REMORA_EXTRACT_ALIGN_PROFILE_FILE")

SIG_PCTL_RANGE = (0.2, 99.8)
SAMPLE_COLORS = ["k", "tab:red", "tab:cyan", "tab:pink", "tab:orange"]
BASE_COLORS = {
    "A": "#00CC00",
    "C": "#0000CC",
    "G": "#FFB300",
    "T": "#CC0000",
    "U": "#CC0000",
    "N": "#FFFFFF",
}


##############
# General IO #
##############


@dataclass
class RefRegion:
    ctg: str
    strand: str
    start: int
    end: int = None

    @property
    def len(self):
        if self.end is None:
            return 1
        return self.end - self.start

    @classmethod
    def parse_ref_region_str(cls, ref_reg_str, req_strand=True):
        mat = re.match(
            r"^(?P<ctg>.+):(?P<st>\d+)-(?P<en>\d+):(?P<strand>[\+\-])$"
            if req_strand
            else r"^(?P<ctg>.+):(?P<st>\d+)-(?P<en>\d+)(:(?P<strand>[\+\-]))?$",
            ref_reg_str,
        )
        if mat is None:
            raise RemoraError(f"Invalid reference region: {ref_reg_str}")
        start = int(mat.group("st")) - 1
        if start < 0:
            raise RemoraError("Invalid reference start coordinate")
        return cls(
            ctg=mat.group("ctg"),
            strand=mat.group("strand"),
            start=start,
            end=int(mat.group("en")),
        )

    @property
    def coord_range(self):
        return range(self.start, self.end)

    def adjust(self, start_adjust=0, end_adjust=0, ref_orient=True):
        """Return a copy of this region adjusted by the specified amounts.

        Args:
            start_adjust (int): Adjustment to start
            end_adjust (int): Adjustment to end
            ref_orient (bool): If True, expand start and end directly. If False
                (read oriented) and region is reverse strand, swap start and
                end adjustments. For read_orient, the region will be expanded
                relative to the reads mapping to the appropriate strand.
        """
        if ref_orient or self.strand == "+":
            end_coord = None if self.end is None else self.end + end_adjust
            return RefRegion(
                self.ctg, self.strand, self.start + start_adjust, end_coord
            )
        else:
            end_coord = None if self.end is None else self.end - start_adjust
            return RefRegion(
                self.ctg, self.strand, self.start - end_adjust, end_coord
            )


def parse_bed_lines(bed_path):
    with open(bed_path) as regs_fh:
        for line in regs_fh:
            fields = line.split()
            ctg, st, en = fields[:3]
            st = int(st)
            en = int(en)
            strand = (
                None if len(fields) < 6 or fields[5] not in "+-" else fields[5]
            )
            yield RefRegion(ctg, strand, st, en)


def parse_bed(bed_path):
    regs = defaultdict(set)
    for ref_reg in parse_bed_lines(bed_path):
        if ref_reg.strand is None:
            for strand in "+-":
                regs[(ref_reg.ctg, strand)].update(ref_reg.coord_range)
        else:
            regs[(ref_reg.ctg, ref_reg.strand)].update(ref_reg.coord_range)
    return regs


def parse_mods_bed(bed_path):
    regs = defaultdict(dict)
    all_mods = set()
    with open(bed_path) as regs_fh:
        for line in regs_fh:
            fields = line.split()
            ctg, st, en, mod = fields[:4]
            all_mods.update(mod)
            if len(fields) < 6 or fields[5] not in "+-":
                for strand in "+-":
                    for pos in range(int(st), int(en)):
                        regs[(ctg, strand)][pos] = mod
            else:
                for pos in range(int(st), int(en)):
                    regs[(ctg, fields[5])][pos] = mod
    return regs, all_mods


def read_is_primary(read):
    """Helper function to return whether a read is a primary mapping
    (not supplementary or secondary)

    Args:
        read: pysam.AlignedSegment
    """
    return not (read.is_supplementary or read.is_secondary)


def strands_match(strand, bam_read):
    return (
        strand not in "+-"
        or (strand == "+" and bam_read.is_forward)
        or (strand == "-" and bam_read.is_reverse)
    )


@dataclass
class ReadIndexedBam:
    """Index bam file by read id. Note that the BAM file handle is closed after
    initialization. Any other operation (e.g. fetch, get_alignments,
    get_first_alignment) will open the pysam file handle and leave it open.
    This allows easier use with multiprocessing using standard operations.

    Args:
        bam_path (str): Path to BAM file
        skip_non_primary (bool): Should non-primary alignmets be skipped
        req_tags (bool): Skip reads without required tags
        read_id_converter (Callable[[str], str]): Function to convert read ids
            (e.g. for concatenated duplex read ids)
    """

    bam_path: str
    skip_non_primary: bool = True
    req_tags: set = None
    read_id_converter: Callable = None

    def __post_init__(self):
        self.num_reads = None
        self.bam_fh = None
        self._bam_idx = None
        self.compute_read_index()

    def open(self):
        # hide warnings for no index when using unmapped or unsorted files
        self.pysam_save = pysam.set_verbosity(0)
        self.bam_fh = pysam.AlignmentFile(
            self.bam_path, mode="rb", check_sq=False
        )

    def close(self):
        self.bam_fh.close()
        self.bam_fh = None
        pysam.set_verbosity(self.pysam_save)

    def fetch(self, ref_reg):
        if self.bam_fh is None:
            self.open()
        for read in self.bam_fh.fetch(ref_reg.ctg, ref_reg.start, ref_reg.end):
            if strands_match(ref_reg.strand, read):
                yield read

    def compute_read_index(self):
        bam_was_closed = self.bam_fh is None
        if bam_was_closed:
            self.open()
        self._bam_idx = defaultdict(list)
        pbar = tqdm(
            smoothing=0,
            unit=" Reads",
            desc="Indexing BAM by read id",
            disable=os.environ.get("LOG_SAFE", False),
        )
        self.num_records = 0
        # iterating over file handle gives incorrect pointers
        while True:
            read_ptr = self.bam_fh.tell()
            try:
                read = next(self.bam_fh)
            except StopIteration:
                break
            pbar.update()
            if self.req_tags is not None:
                tags = set(tg[0] for tg in read.tags)
                if not self.req_tags.issubset(tags):
                    LOGGER.debug(
                        f"{read.query_name} missing tags "
                        f"{self.req_tags.difference(tags)}"
                    )
                    continue
            index_read_id = (
                read.query_name
                if self.read_id_converter is None
                else self.read_id_converter(read.query_name)
            )
            if self.skip_non_primary and (
                not read_is_primary(read) or index_read_id in self._bam_idx
            ):
                LOGGER.debug(f"{read.query_name} not primary")
                continue
            self.num_records += 1
            self._bam_idx[index_read_id].append(read_ptr)
        # close bam if it was closed at start of function call
        if bam_was_closed:
            self.close()
        pbar.close()
        # convert defaultdict to dict
        self._bam_idx = dict(self._bam_idx)
        self.num_reads = len(self._bam_idx)

    def get_alignments(self, read_id):
        if self._bam_idx is None:
            raise RemoraError("Bam index not yet computed")
        if self.bam_fh is None:
            self.open()
        try:
            read_ptrs = self._bam_idx[read_id]
        except KeyError:
            raise RemoraError(f"Could not find {read_id} in {self.bam_path}")
        for read_ptr in read_ptrs:
            self.bam_fh.seek(read_ptr)
            try:
                bam_read = next(self.bam_fh)
            except OSError:
                LOGGER.debug(
                    f"Could not extract {read_id} from {self.bam_path} "
                    f"at {read_ptr}"
                )
                raise RemoraError(
                    "Could not extract BAM read. Ensure BAM file object was "
                    "closed before spawning process."
                )
            assert bam_read.query_name == read_id, (
                f"Read ID {read_id} does not match extracted BAM record "
                f"{bam_read.query_name} at {read_ptr}. ReadIndexedBam may "
                "be corrupted."
            )
            yield bam_read

    def get_first_alignment(self, read_id):
        return next(self.get_alignments(read_id))

    def __contains__(self, read_id):
        assert isinstance(read_id, str)
        return read_id in self._bam_idx

    def __getitem__(self, read_id):
        return self._bam_idx[read_id]

    def __del__(self):
        if self.bam_fh is not None:
            self.bam_fh.close()

    @cached_property
    def read_ids(self):
        return list(self._bam_idx.keys())


def parse_move_tag(
    mv_tag, sig_len, seq_len=None, check=True, reverse_signal=False
):
    stride = mv_tag[0]
    mv_table = np.array(mv_tag[1:])
    query_to_signal = np.nonzero(mv_table)[0] * stride
    query_to_signal = np.concatenate([query_to_signal, [sig_len]])
    if reverse_signal:
        query_to_signal = sig_len - query_to_signal[::-1]
    if check and seq_len is not None and query_to_signal.size - 1 != seq_len:
        LOGGER.debug(
            f"Move table (num moves: {query_to_signal.size - 1}) discordant "
            f"with basecalls (seq len: {seq_len})"
        )
        raise RemoraError("Move table discordant with basecalls")
    if check and mv_table.size != sig_len // stride:
        LOGGER.debug(
            f"Move table (len: {mv_table.size}) discordant with "
            f"signal (sig len // stride: {sig_len // stride})"
        )
        raise RemoraError("Move table discordant with signal")
    return query_to_signal, mv_table, stride


#######################
# POD5/BAM Extraction #
#######################


def iter_pod5_reads(pod5_path, num_reads=None, read_ids=None):
    """Iterate over Pod5Read objects

    Args:
        pod5_path (str): Path to POD5 file
        num_reads (int): Maximum number of reads to iterate
        read_ids (iterable): Read IDs to extract

    Yields:
        Requested pod5.reader.ReadRecord objects
    """
    LOGGER.debug(f"Reading from POD5 at {pod5_path}")
    with pod5.Reader(Path(pod5_path)) as pod5_fh:
        for read_num, read in enumerate(
            pod5_fh.reads(selection=read_ids, preload=["samples"])
        ):
            if num_reads is not None and read_num >= num_reads:
                LOGGER.debug(
                    f"Completed pod5 signal worker, reached {num_reads}."
                )
                return

            yield read
    LOGGER.debug("Completed pod5 signal worker")


def iter_signal(pod5_path, num_reads=None, read_ids=None, rev_sig=False):
    """Iterate io Read objects loaded from Pod5

    Args:
        pod5_path (str): Path to POD5 file
        num_reads (int): Maximum number of reads to iterate
        read_ids (iterable): Read IDs to extract
        rev_sig (bool): Should signal be reversed on reading

    Yields:
        2-tuple:
            1. remora.io.Read object
            2. Error text or None if no errors
    """
    for pod5_read in iter_pod5_reads(
        pod5_path=pod5_path, num_reads=num_reads, read_ids=read_ids
    ):
        dacs = pod5_read.signal[::-1] if rev_sig else pod5_read.signal
        read = Read(
            read_id=str(pod5_read.read_id),
            dacs=dacs,
            shift_dacs_to_pa=pod5_read.calibration.offset,
            scale_dacs_to_pa=pod5_read.calibration.scale,
        )

        yield read, None
    LOGGER.debug("Completed signal worker")


if _SIG_PROF_FN:
    _iter_signal_wrapper = iter_signal

    def iter_signal(*args, **kwargs):
        import cProfile

        sig_prof = cProfile.Profile()
        retval = sig_prof.runcall(_iter_signal_wrapper, *args, **kwargs)
        sig_prof.dump_stats(_SIG_PROF_FN)
        return retval


def extract_alignments(read_err, bam_idx, rev_sig=False):
    io_read, err = read_err
    if io_read is None:
        return [read_err]
    read_alignments = []
    try:
        for bam_read in bam_idx.get_alignments(io_read.read_id):
            align_read = io_read.copy()
            align_read.add_alignment(bam_read, reverse_signal=rev_sig)
            read_alignments.append(tuple((align_read, None)))
    except KeyError:
        return [tuple((None, "Read id not found in BAM file"))]
    except RemoraError as e:
        return [tuple((None, str(e)))]
    return read_alignments


if _ALIGN_PROF_FN:
    _extract_align_wrapper = extract_alignments

    def extract_alignments(*args, **kwargs):
        import cProfile

        align_prof = cProfile.Profile()
        retval = align_prof.runcall(_extract_align_wrapper, *args, **kwargs)
        align_prof.dump_stats(_ALIGN_PROF_FN)
        return retval


def iter_regions(bam_fh, reg_len=100_000):
    for ctg, ctg_len in zip(bam_fh.header.references, bam_fh.header.lengths):
        for st in range((ctg_len // reg_len) + 1):
            yield RefRegion(
                ctg=ctg,
                strand="+",
                start=st * reg_len,
                end=(st + 1) * reg_len,
            )
            yield RefRegion(
                ctg=ctg,
                strand="-",
                start=st * reg_len,
                end=(st + 1) * reg_len,
            )


def get_reg_bam_reads(ref_reg, bam_fh):
    return [
        bam_read
        for bam_read in bam_fh.fetch(ref_reg.ctg, ref_reg.start, ref_reg.end)
        if read_is_primary(bam_read) and strands_match(ref_reg.strand, bam_read)
    ]


def iter_covered_regions(
    bam_path, chunk_len=1_000, max_chunk_cov=None, pickle_safe=False
):
    with pysam.AlignmentFile(bam_path) as bam_fh:
        for reg in iter_regions(bam_fh, chunk_len):
            bam_reads = get_reg_bam_reads(reg, bam_fh)
            if len(bam_reads) == 0:
                continue
            if max_chunk_cov is not None:
                target_bases = chunk_len * max_chunk_cov
                total_bases = 0
                random.shuffle(bam_reads)
                sampled_bam_reads = []
                for bam_read in bam_reads:
                    sampled_bam_reads.append(bam_read)
                    total_bases += min(bam_read.reference_end, reg.end) - max(
                        bam_read.reference_start, reg.start
                    )
                    if total_bases >= target_bases:
                        break
                bam_reads = sampled_bam_reads
            if pickle_safe:
                yield reg, [br.to_dict() for br in bam_reads]
                continue
            yield reg, bam_reads


def compute_base_space_sig_coords(seq_to_sig_map):
    """Compute coordinates for signal points, interpolating signal assigned to
    each base linearly through the span of each covered base.
    """
    return np.interp(
        np.arange(seq_to_sig_map[-1] - seq_to_sig_map[0]),
        seq_to_sig_map,
        np.arange(seq_to_sig_map.size),
    )


@dataclass
class ReadRefReg:
    read_id: str
    norm_signal: np.ndarray
    seq: str
    seq_to_sig_map: np.ndarray
    ref_reg: RefRegion

    @property
    def ref_sig_coords(self):
        """Compute signal coorindates for plotting this read against the
        mapped stretch of reference.
        """
        return (
            compute_base_space_sig_coords(self.seq_to_sig_map)
            + self.ref_reg.start
        )

    def plot_on_signal_coords(self, **kwargs):
        """Plot signal on signal coordinates. See global plot_on_signal_coords
        function for kwargs.
        """
        return plot_on_signal_coords(
            self.seq,
            self.norm_signal,
            self.seq_to_sig_map,
            rev_strand=self.ref_reg.strand == "-",
            **kwargs,
        )

    def plot_on_base_coords(self, **kwargs):
        """Plot signal on base/sequence coordinates. See global
        plot_on_base_coords function for kwargs.
        """
        return plot_on_base_coords(
            self.ref_reg.start,
            self.seq,
            self.norm_signal,
            self.seq_to_sig_map,
            rev_strand=self.ref_reg.strand == "-",
            **kwargs,
        )


@dataclass
class ReadBasecallRegion:
    read_id: str
    norm_signal: np.ndarray
    seq: str
    seq_to_sig_map: np.ndarray
    start: int

    def plot_on_signal_coords(self, **kwargs):
        """Plot signal on signal coordinates. See global plot_on_signal_coords
        function for kwargs.
        """
        return plot_on_signal_coords(
            self.seq, self.norm_signal, self.seq_to_sig_map, **kwargs
        )

    def plot_on_base_coords(self, **kwargs):
        """Plot signal on base/sequence coordinates. See global
        plot_on_base_coords function for kwargs.
        """
        return plot_on_base_coords(
            self.start,
            self.seq,
            self.norm_signal,
            self.seq_to_sig_map,
            **kwargs,
        )


def get_ref_int_seq_from_reads(ref_reg, bam_reads, ref_orient=True):
    # fill with -2 since N is represented by -1
    int_seq = np.full(ref_reg.len, -2, np.int32)
    # extract forward reference sequence.
    for bam_read in bam_reads:
        read_ref_seq = bam_read.get_reference_sequence().upper()
        int_seq[
            max(0, bam_read.reference_start - ref_reg.start) : (
                bam_read.reference_end - ref_reg.start
            )
        ] = util.seq_to_int(
            read_ref_seq[
                max(0, ref_reg.start - bam_read.reference_start) : (
                    ref_reg.end - bam_read.reference_start
                )
            ]
        )
        if not np.any(int_seq == -2):
            break
    if ref_reg.strand == "-":
        if ref_orient:
            return util.comp_np(int_seq)
        else:
            return util.revcomp_np(int_seq)
    return int_seq


def get_ref_seq_from_reads(ref_reg, bam_reads, ref_orient=True):
    int_seq = get_ref_int_seq_from_reads(
        ref_reg, bam_reads, ref_orient=ref_orient
    )
    int_seq[np.equal(int_seq, -2)] = -1
    return util.int_to_seq(int_seq)


def get_ref_seq_and_levels_from_reads(
    ref_reg,
    bam_reads,
    sig_map_refiner,
    ref_orient=True,
):
    """Extract sequence and levels from BAM reads covering a region.

    Args:
        ref_reg (RefRegion): Reference region
        bam_reads (iterable): pysam.AlignedSegments covering ref_reg
        sig_map_refiner (SigMapRefiner): For level extraction
        ref_orient (bool): Should returned sequence and levels be reference
            oriented? This only effects the output for reverse strand ref_reg.
            Reference orientation will return the reference sequence and levels
            in the forward reference direction. Read orient (ref_orient=False)
            will return sequence and levels in 5' to 3' read direction on the
            reverse strand.
    """
    # extract read oriented context seq
    context_int_seq = get_ref_int_seq_from_reads(
        ref_reg.adjust(
            -sig_map_refiner.bases_before,
            sig_map_refiner.bases_after,
            ref_orient=False,
        ),
        bam_reads,
        ref_orient=False,
    )
    # levels are read oriented
    levels = sig_map_refiner.extract_levels(context_int_seq)
    # convert sequence positions not in reads to nan and seq to N
    levels[np.equal(context_int_seq, -2)] = np.NAN
    context_int_seq[np.equal(context_int_seq, -2)] = -1
    seq = util.int_to_seq(context_int_seq)
    # trim seq and levels
    seq = seq[
        sig_map_refiner.bases_before : sig_map_refiner.bases_before
        + ref_reg.len
    ]
    levels = levels[
        sig_map_refiner.bases_before : sig_map_refiner.bases_before
        + ref_reg.len
    ]

    if ref_reg.strand == "-" and ref_orient:
        seq = seq[::-1]
        levels = levels[::-1]
    return seq, levels


def get_pod5_reads(pod5_fh, read_ids):
    return dict(
        (str(pod5_read.read_id), pod5_read)
        for pod5_read in pod5_fh.reads(selection=read_ids, preload=["samples"])
    )


def get_io_reads(bam_reads, pod5_fh, reverse_signal=False, missing_ok=False):
    pod5_reads = get_pod5_reads(
        pod5_fh, [bam_read.query_name for bam_read in bam_reads]
    )
    io_reads = []
    for bam_read in bam_reads:
        try:
            io_read = Read.from_pod5_and_alignment(
                pod5_read_record=pod5_reads[bam_read.query_name],
                alignment_record=bam_read,
                reverse_signal=reverse_signal,
            )
        except Exception:
            if missing_ok:
                continue
            else:
                raise RemoraError("BAM record not found in POD5")
        io_reads.append(io_read)
    return io_reads


def get_reads_reference_regions(
    ref_reg,
    pod5_and_bam_fhs,
    sig_map_refiner=None,
    skip_sig_map_refine=False,
    max_reads=50,
    reverse_signal=False,
):
    all_bam_reads = []
    samples_read_ref_regs = []
    for pod5_fh, bam_fh in pod5_and_bam_fhs:
        sample_bam_reads = get_reg_bam_reads(ref_reg, bam_fh)
        if len(sample_bam_reads) == 0:
            raise RemoraError("No reads covering region")
        if max_reads is not None and len(sample_bam_reads) > max_reads:
            sample_bam_reads = random.sample(sample_bam_reads, max_reads)
        all_bam_reads.append(sample_bam_reads)
        io_reads = get_io_reads(sample_bam_reads, pod5_fh, reverse_signal)
        if sig_map_refiner is not None and not skip_sig_map_refine:
            for io_read in io_reads:
                io_read.set_refine_signal_mapping(
                    sig_map_refiner, ref_mapping=True
                )
        samples_read_ref_regs.append(
            [io_read.extract_ref_reg(ref_reg) for io_read in io_reads]
        )
    return samples_read_ref_regs, all_bam_reads


def get_ref_reg_sample_metrics(
    ref_reg,
    pod5_fh,
    bam_reads,
    metric,
    sig_map_refiner,
    skip_sig_map_refine=False,
    reverse_signal=False,
    ref_orient=True,
    **kwargs,
):
    io_reads = get_io_reads(bam_reads, pod5_fh, reverse_signal)
    if sig_map_refiner is not None and not skip_sig_map_refine:
        for io_read in io_reads:
            io_read.set_refine_signal_mapping(sig_map_refiner, ref_mapping=True)
    sample_metrics = [
        io_read.compute_per_base_metric(metric, region=ref_reg, **kwargs)
        for io_read in io_reads
    ]
    if len(sample_metrics) > 0:
        reg_metrics = dict(
            (
                metric_name,
                np.stack([mv[metric_name] for mv in sample_metrics]),
            )
            for metric_name in sample_metrics[0].keys()
        )
        # ref_anchored read metrics are read oriented. Thus if ref_orient=True,
        # need to flip metrics
        if ref_orient and ref_reg.strand == "-":
            return dict(
                (metric_name, vals[:, ::-1])
                for metric_name, vals in reg_metrics.items()
            )
        return reg_metrics


def get_ref_reg_samples_metrics(
    ref_reg,
    pod5_and_bam_fhs,
    sig_map_refiner=None,
    skip_sig_map_refine=False,
    max_reads=None,
    reverse_signal=False,
    metric="dwell_trimmean",
    **kwargs,
):
    all_bam_reads = []
    samples_metrics = []
    for pod5_fh, bam_fh in pod5_and_bam_fhs:
        sample_bam_reads = get_reg_bam_reads(ref_reg, bam_fh)
        if len(sample_bam_reads) == 0:
            raise RemoraError("No reads covering region")
        if max_reads is not None and len(sample_bam_reads) > max_reads:
            sample_bam_reads = random.sample(sample_bam_reads, max_reads)
        all_bam_reads.append(sample_bam_reads)
        sample_metrics = get_ref_reg_sample_metrics(
            ref_reg,
            pod5_fh,
            sample_bam_reads,
            metric,
            sig_map_refiner,
            skip_sig_map_refine,
            reverse_signal,
            **kwargs,
        )
        if sample_metrics is not None:
            samples_metrics.append(sample_metrics)

    return samples_metrics, all_bam_reads


###################
# K-mer functions #
###################


def get_region_kmers(
    reg_and_bam_reads,
    pod5_fh,
    sig_map_refiner,
    kmer_context_bases,
    min_cov=10,
    start_trim=2,
    end_trim=2,
    dict_bam_reads=False,
    bam_header=None,
    reverse_signal=False,
):
    reg, bam_reads = reg_and_bam_reads
    if dict_bam_reads:
        bam_reads = [
            pysam.AlignedSegment.from_dict(br, bam_header) for br in bam_reads
        ]
    reg_metrics = get_ref_reg_sample_metrics(
        reg,
        pod5_fh,
        bam_reads,
        "dwell_trimmean",
        sig_map_refiner,
        start_trim=start_trim,
        end_trim=end_trim,
        ref_orient=False,
        reverse_signal=reverse_signal,
    )
    seq = get_ref_seq_from_reads(
        reg.adjust(
            -kmer_context_bases[0], kmer_context_bases[1], ref_orient=False
        ),
        bam_reads,
        ref_orient=False,
    )
    kmer_len = sum(kmer_context_bases) + 1
    reg_kmer_levels = dict(
        ("".join(bs), []) for bs in product("ACGT", repeat=kmer_len)
    )
    for offset in range(reg.len):
        kmer = seq[offset : offset + kmer_len]
        try:
            offset_kmer_levels = reg_kmer_levels[kmer]
        except KeyError:
            continue
        site_read_levels = reg_metrics["trimmean"][:, offset]
        site_read_levels = site_read_levels[np.isfinite(site_read_levels)]
        if site_read_levels.size < min_cov:
            continue
        offset_kmer_levels.append(np.median(site_read_levels))
    return dict(
        (kmer, np.array(levels)) for kmer, levels in reg_kmer_levels.items()
    )


def prep_region_kmers(*args, **kwargs):
    args = list(args)
    args[0] = pod5.Reader(args[0])
    return tuple(args), kwargs


def get_site_kmer_levels(
    pod5_path,
    bam_path,
    sig_map_refiner,
    kmer_context_bases,
    min_cov=10,
    chunk_len=1_000,
    max_chunk_cov=100,
    start_trim=1,
    end_trim=1,
    num_workers=1,
    reverse_signal=False,
):
    regs_bam_reads = util.BackgroundIter(
        iter_covered_regions,
        args=(bam_path, chunk_len, max_chunk_cov),
        kwargs={"pickle_safe": True},
    )
    with pysam.AlignmentFile(bam_path) as bam_fh:
        bam_header = bam_fh.header
    regs_kmer_levels = util.MultitaskMap(
        get_region_kmers,
        regs_bam_reads,
        prep_func=prep_region_kmers,
        num_workers=num_workers,
        use_process=True,
        args=(
            pod5_path,
            sig_map_refiner,
            kmer_context_bases,
        ),
        kwargs={
            "min_cov": min_cov,
            "start_trim": start_trim,
            "end_trim": start_trim,
            "dict_bam_reads": True,
            "bam_header": bam_header,
            "reverse_signal": reverse_signal,
        },
        name="GetKmers",
    )

    # enumerate kmers for dict
    kmer_len = sum(kmer_context_bases) + 1
    all_kmer_levels = dict(
        ("".join(bs), []) for bs in product("ACGT", repeat=kmer_len)
    )
    for reg_kmer_levels in regs_kmer_levels:
        for kmer, levels in reg_kmer_levels.items():
            all_kmer_levels[kmer].append(levels)
    return dict(
        (kmer, np.concatenate(levels) if len(levels) > 0 else np.array([]))
        for kmer, levels in all_kmer_levels.items()
    )


######################
# Plotting functions #
######################


def plot_on_signal_coords(
    seq,
    sig,
    seq_to_sig_map,
    rev_strand=False,
    levels=None,
    fig_ax=None,
    figsize=DEFAULT_PLOT_FIG_SIZE,
    sig_lw=8,
    levels_lw=8,
    ylim=None,
    t_as_u=False,
):
    """Plot a single read on signal coordinates.

    Args:
        seq (str): Sequence to plot
        sig (np.array): Signal to plot
        seq_to_sig_map (np.array): Mapping from sequence to signal coordinates
        rev_strand (bool): Plot seq with 180 rotation
        levels (np.array): Expected signal levels for bases in region
        fig_ax (tuple): If None, new figure/axes will be opened
        figsize (tuple): option to pass to plt.subplots if ax is None
        sig_lw (int): Linewidth for signal lines
        levels_lw (int): Linewidth for level lines (if applicable)
        ylim (tuple): 2-tuple with y-axis limits
        t_as_u (bool): Plot T bases as U (RNA)

    Returns:
        matplotlib axis
    """
    if fig_ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig, ax = fig_ax
    if ylim is None:
        sig_min, sig_max = np.percentile(sig, SIG_PCTL_RANGE)
        sig_diff = sig_max - sig_min
        ylim = (sig_min - sig_diff * 0.01, sig_max + sig_diff * 0.01)
    base_text_loc = ylim[0] + ((ylim[1] - ylim[0]) * 0.02)

    # plot vertical base lines
    for b in seq_to_sig_map:
        ax.axvline(x=b, color="k", alpha=0.1, lw=1)
    # plot read signal
    ax.plot(
        np.arange(seq_to_sig_map[0], seq_to_sig_map[-1]),
        sig,
        color="k",
        alpha=0.5,
        lw=sig_lw,
    )
    # plot levels
    if levels is not None:
        for b_lev, b_st, b_en in zip(
            levels, seq_to_sig_map[:-1], seq_to_sig_map[1:]
        ):
            ax.plot(
                [b_st, b_en],
                [b_lev] * 2,
                linestyle="-",
                color="y",
                alpha=0.5,
                lw=levels_lw,
            )
    # plot bases
    if t_as_u:
        seq = util.t_to_u(seq)
    for b_st, b_en, base in zip(seq_to_sig_map[:-1], seq_to_sig_map[1:], seq):
        ax.text(
            (b_en + b_st) / 2,
            base_text_loc,
            base,
            color=BASE_COLORS[base],
            ha="center",
            size=30,
            rotation=180 if rev_strand else 0,
        )
    ax.set_ylim(*ylim)
    ax.set_ylabel("Normalized Signal", fontsize=45)
    ax.set_xlabel("Signal Position", fontsize=45)
    ax.tick_params(labelsize=36)
    return fig, ax


def plot_on_base_coords(
    start_base,
    seq,
    sig,
    seq_to_sig_map,
    levels=None,
    rev_strand=False,
    fig_ax=None,
    figsize=DEFAULT_PLOT_FIG_SIZE,
    sig_lw=8,
    levels_lw=8,
    ylim=None,
    t_as_u=False,
):
    """Plot a single read on base/sequence coordinates.

    Args:
        start_base (int): Coordinate of first base in seq
        seq (str): Sequence to plot
        sig (np.array): Signal to plot
        seq_to_sig_map (np.array): Mapping from sequence to signal coordinates
        levels (np.array): Expected signal levels for bases in region
        rev_strand (bool): Plot bases upside down
        fig_ax (tuple): If None, new figure/axes will be opened
        figsize (tuple): option to pass to plt.subplots if ax is None
        sig_lw (int): Linewidth for signal lines
        levels_lw (int): Linewidth for level lines (if applicable)
        ylim (tuple): 2-tuple with y-axis limits
        t_as_u (bool): Plot T bases as U (RNA)

    Returns:
        matplotlib axis
    """
    if fig_ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig, ax = fig_ax
    if ylim is None:
        sig_min, sig_max = np.percentile(sig, SIG_PCTL_RANGE)
        sig_diff = sig_max - sig_min
        ylim = (sig_min - sig_diff * 0.01, sig_max + sig_diff * 0.01)
    base_text_loc = ylim[0] + ((ylim[1] - ylim[0]) * 0.02)

    # plot vertical base lines
    for b in np.arange(start_base, start_base + seq_to_sig_map.size):
        ax.axvline(x=b, color="k", alpha=0.1, lw=1)
    # plot read signal
    ax.plot(
        compute_base_space_sig_coords(seq_to_sig_map) + start_base,
        sig,
        color="k",
        alpha=0.5,
        lw=sig_lw,
    )
    # plot levels
    if levels is not None:
        for b_num, b_lev in enumerate(levels):
            ax.plot(
                [start_base + b_num, start_base + b_num + 1],
                [b_lev] * 2,
                linestyle="-",
                color="y",
                alpha=0.5,
                lw=levels_lw,
            )
    # plot bases
    if t_as_u:
        seq = util.t_to_u(seq)
    for b_num, base in enumerate(seq):
        ax.text(
            start_base + b_num + 0.5,
            base_text_loc,
            base,
            color=BASE_COLORS[base],
            ha="center",
            size=30,
            rotation=180 if rev_strand else 0,
        )
    ax.set_ylim(*ylim)
    ax.set_ylabel("Normalized Signal", fontsize=45)
    ax.set_xlabel("Base Position", fontsize=45)
    ax.tick_params(labelsize=36)
    return fig, ax


def plot_ref_region_reads(
    ref_reg,
    ref_reg_reads,
    seq,
    levels,
    max_reads=50,
    fig_ax=None,
    figsize=DEFAULT_PLOT_FIG_SIZE,
    sig_lw=2,
    sig_cols=SAMPLE_COLORS,
    levels_lw=6,
    ylim=None,
    highlight_ranges=None,
    t_as_u=False,
):
    # start plotting
    if fig_ax is None:
        fig, ax = plt.subplots(figsize=figsize)
    else:
        fig, ax = fig_ax

    # plot highlight regions
    if highlight_ranges is not None:
        for st, en, col in highlight_ranges:
            ax.axvspan(st, en, facecolor=col, alpha=0.4)
    # plot vertical base lines
    for b in np.arange(ref_reg.start, ref_reg.end + 1):
        ax.axvline(x=b, color="k", alpha=0.1, lw=1)
    # plot read signal
    for sample_col, sample_reads in zip(sig_cols, ref_reg_reads):
        for read_reg in sample_reads:
            ax.plot(
                read_reg.ref_sig_coords,
                read_reg.norm_signal,
                color=sample_col,
                alpha=0.1,
                lw=sig_lw,
            )
    # plot levels
    if levels is not None:
        for b_num, b_lev in enumerate(levels):
            ax.plot(
                [ref_reg.start + b_num, ref_reg.start + b_num + 1],
                [b_lev] * 2,
                linestyle="-",
                color="y",
                alpha=0.5,
                lw=levels_lw,
            )
    # plot bases
    if ylim is None:
        sig_min, sig_max = np.percentile(
            np.concatenate([r.norm_signal for r in chain(*ref_reg_reads)]),
            SIG_PCTL_RANGE,
        )
        sig_diff = sig_max - sig_min
        ylim = (sig_min - sig_diff * 0.01, sig_max + sig_diff * 0.01)
    base_text_loc = ylim[0] + ((ylim[1] - ylim[0]) * 0.02)
    rotation = 0 if ref_reg.strand == "+" else 180
    if t_as_u:
        seq = util.t_to_u(seq)
    for b_num, base in enumerate(seq):
        ax.text(
            ref_reg.start + b_num + 0.5,
            base_text_loc,
            base,
            color=BASE_COLORS[base],
            ha="center",
            size=30,
            rotation=rotation,
        )
    ax.set_ylim(*ylim)
    ax.set_xlim(ref_reg.start, ref_reg.end)
    ax.set_title(ref_reg.ctg, fontsize=50)
    ax.set_ylabel("Normalized Signal", fontsize=45)
    ax.set_xlabel("Reference Position", fontsize=45)
    ax.tick_params(labelsize=36)
    # shift tick labels left by 0.5 plot units to match genome browsers
    xticks = np.array(
        [int(xt) for xt in ax.get_xticks() if ref_reg.start < xt < ref_reg.end]
    )
    ax.set_xticks(xticks - 0.5)
    ax.set_xticklabels(xticks)
    return fig, ax


def plot_signal_at_ref_region(
    ref_reg,
    pod5_and_bam_fhs,
    sig_map_refiner=None,
    skip_sig_map_refine=False,
    max_reads=50,
    reverse_signal=False,
    t_as_u=False,
    fig_ax=None,
    figsize=DEFAULT_PLOT_FIG_SIZE,
    sig_lw=2,
    levels_lw=6,
    ylim=None,
    highlight_ranges=None,
):
    """Plot signal from reads at a reference region.

    Args:
        bam_fhs (pysam.AlignmentFile): Sorted and indexed BAM file handle or
            a list of them
        pod5_fhs (str): POD5 file handles or a list of them
        ref_reg (RefRegion): Reference position at which to plot signal
        sig_map_refiner (SigMapRefiner): For signal mapping and level extract
        skip_sig_map_refine (bool): Skip signal mapping refinement
        max_reads (int): Maximum reads to plot (TODO: add overplotting options)
        reverse_signal (bool): Is nanopore signal 3'>5' orientation?
        t_as_u (bool): Plot T bases as U (RNA)
        fig_ax (tuple): If None, new figure/axes will be opened
        figsize (tuple): option to pass to plt.subplots if ax is None
        sig_lw (int): Linewidth for signal lines
        levels_lw (int): Linewidth for level lines (if applicable)
        ylim (tuple): 2-tuple with y-axis limits
        highlight_ranges (iterable): Elements should be 3-tuples containing
            reference start, end and color values.

    Returns:
        matplotlib axis
    """
    read_ref_regs, reg_bam_reads = get_reads_reference_regions(
        ref_reg,
        pod5_and_bam_fhs,
        sig_map_refiner=sig_map_refiner,
        skip_sig_map_refine=skip_sig_map_refine,
        max_reads=max_reads,
        reverse_signal=reverse_signal,
    )
    seq, levels = get_ref_seq_and_levels_from_reads(
        ref_reg, chain(*reg_bam_reads), sig_map_refiner
    )
    fig_ax = plot_ref_region_reads(
        ref_reg,
        read_ref_regs,
        seq,
        levels,
        fig_ax=fig_ax,
        figsize=figsize,
        sig_lw=sig_lw,
        levels_lw=levels_lw,
        ylim=ylim,
        highlight_ranges=highlight_ranges,
        t_as_u=t_as_u,
    )
    return fig_ax


def plot_ref_region_metrics(
    ref_reg,
    samples_metrics,
    fig_axs=None,
    fig_width=DEFAULT_PLOT_FIG_SIZE[0],
    facet_height=DEFAULT_PLOT_FIG_SIZE[1],
    highlight_ranges=None,
    sample_names=None,
    t_as_u=False,
):
    metric_names = list(samples_metrics[0].keys())
    if fig_axs is None:
        fig, axs = plt.subplots(
            len(metric_names),
            figsize=(fig_width, facet_height * len(metric_names)),
            sharex=True,
        )
    else:
        fig, axs = fig_axs
    if sample_names is None:
        sample_names = [f"Sample{idx}" for idx in range(len(samples_metrics))]

    for ax, metric_name in zip(axs, metric_names):
        samples_num_reads = np.array(
            [sm[metric_name].shape[0] for sm in samples_metrics]
        )
        plot_data = pd.DataFrame(
            {
                metric_name: np.concatenate(
                    [sm[metric_name].flatten() for sm in samples_metrics]
                ),
                "Position": np.tile(
                    np.arange(ref_reg.len), [samples_num_reads.sum()]
                ),
                "Sample": np.repeat(
                    sample_names, samples_num_reads * ref_reg.len
                ),
            }
        )
        plot_data = plot_data.loc[~np.isnan(plot_data[metric_name])]

        with np.errstate(invalid="ignore"):
            _ = sns.boxplot(
                data=plot_data,
                x="Position",
                y=metric_name,
                hue="Sample",
                ax=ax,
            )
        ax.set_xlabel("", fontsize=45)
        if metric_name == "dwell":
            ax.set_ylim(2, 80)
            ax.set_yscale("log")
        if metric_name.endswith("mean"):
            ax.set_ylim(-1.5, 2)
        if metric_name.endswith("sd"):
            ax.set_ylim(-0.005, 0.3)
            ax.set_xlabel("Position", fontsize=45)
        ax.set_ylabel(metric_name, fontsize=45)
        ax.tick_params(labelsize=16, axis="x")
        ax.tick_params(labelsize=32, axis="y")
    return fig, axs


def plot_metric_at_ref_region(
    ref_reg,
    pod5_and_bam_fhs,
    metric="dwell_trimmean",
    sig_map_refiner=None,
    skip_sig_map_refine=False,
    max_reads=None,
    reverse_signal=False,
    t_as_u=False,
    fig_axs=None,
    fig_width=DEFAULT_PLOT_FIG_SIZE[0],
    facet_height=DEFAULT_PLOT_FIG_SIZE[1],
    highlight_ranges=None,
    **kwargs,
):
    """Plot signal from reads at a reference region.

    Args:
        ref_reg (RefRegion): Reference position to plot
        pod5_and_bam_fhs (list): List of 2-tuples containing sorted and indexed
            BAM file handles and POD5 file handles
        metric (str): Named metric (e.g. dwell, mean, sd). Should be a key
            in metrics.METRIC_FUNCS
        sig_map_refiner (SigMapRefiner): For signal mapping and level extract
        skip_sig_map_refine (bool): Skip signal mapping refinement
        max_reads (int): Maximum reads to plot
        reverse_signal (bool): Is nanopore signal 3'>5' orientation?
        t_as_u (bool): Plot T bases as U (RNA)
        fig_axs (tuple): If None, new figure/axes will be opened. Should be
            same length as the number of metrics computed.
        fig_width (float): Figure width
        facet_height (float): Height of each facet (one per metric)
        highlight_ranges (iterable): Elements should be 3-tuples containing
            reference start, end and color values.

    Returns:
        matplotlib axis
    """
    samples_metrics, all_bam_reads = get_ref_reg_samples_metrics(
        ref_reg,
        pod5_and_bam_fhs,
        metric=metric,
        sig_map_refiner=sig_map_refiner,
        skip_sig_map_refine=skip_sig_map_refine,
        max_reads=max_reads,
        reverse_signal=reverse_signal,
        **kwargs,
    )
    fig_axs = plot_ref_region_metrics(
        ref_reg,
        samples_metrics,
        fig_axs=fig_axs,
        fig_width=fig_width,
        facet_height=facet_height,
        highlight_ranges=highlight_ranges,
        t_as_u=t_as_u,
    )
    plt.tight_layout()
    return fig_axs


###########
# IO Read #
###########


@dataclass
class Read:
    read_id: str
    dacs: np.ndarray = None
    seq: str = None
    stride: int = None
    num_trimmed: int = None
    mv_table: np.array = None
    query_to_signal: np.ndarray = None
    shift_dacs_to_pa: float = None
    scale_dacs_to_pa: float = None
    shift_pa_to_norm: float = None
    scale_pa_to_norm: float = None
    shift_dacs_to_norm: float = None
    scale_dacs_to_norm: float = None
    ref_seq: str = None
    ref_reg: RefRegion = None
    cigar: list = None
    ref_to_signal: np.ndarray = None
    full_align: str = None

    @property
    def pa_signal(self):
        """Picoampere convereted signal. Shift and scale values are determined
        from the instrument. These values are generally not recommended for any
        type of analysis.
        """
        assert self.scale_dacs_to_pa is not None
        assert self.shift_dacs_to_pa is not None
        return self.scale_dacs_to_pa * (self.dacs + self.shift_dacs_to_pa)

    @property
    def norm_signal(self):
        """Normalized signal, generally aiming to standardize the signal with
        mean=0 and SD=1, though different methods may be applied to perform
        this operation robustly.
        """
        assert self.scale_dacs_to_norm is not None
        assert self.shift_dacs_to_norm is not None
        return (self.dacs - self.shift_dacs_to_norm) / self.scale_dacs_to_norm

    def compute_pa_to_norm_scaling(self, factor=PA_TO_NORM_SCALING_FACTOR):
        self.shift_pa_to_norm = np.median(self.pa_signal)
        self.scale_pa_to_norm = max(
            1.0,
            np.median(np.abs(self.pa_signal - self.shift_pa_to_norm)) * factor,
        )

    @property
    def seq_len(self):
        if self.query_to_signal is None:
            return None
        return self.query_to_signal.size - 1

    @property
    def ref_seq_len(self):
        if self.ref_to_signal is None:
            return None
        return self.ref_to_signal.size - 1

    def with_duplex_alignment(self, duplex_read_alignment, duplex_orientation):
        """Return copy with alignment to duplex sequence

        Args:
            duplex_read_alignment (AlignedSegment): pysam alignment record
                containing duplex sequence
            duplex_orientation (bool): Is duplex mapping in the same
                orientation as this simplex read.
        """
        assert self.query_to_signal is not None, "requires query_to_signal"
        assert (
            duplex_read_alignment.query_sequence is not None
        ), "no duplex base call sequence?"
        assert (
            len(duplex_read_alignment.query_sequence) > 0
        ), "duplex base call sequence is empty string?"

        read = copy(self)

        duplex_read_sequence = (
            duplex_read_alignment.query_sequence
            if duplex_orientation
            else util.revcomp(duplex_read_alignment.query_sequence)
        )

        # we don't have a mapping of each base in the simplex sequence to each
        # base in the duplex sequence, so we infer it by alignment. Using the
        # simplex sequence as the query sequence is somewhat arbitrary, but it
        # makes the downstream coordinate mappings more convenient
        simplex_duplex_mapping = DU.map_simplex_to_duplex(
            simplex_seq=read.seq, duplex_seq=duplex_read_sequence
        )
        # duplex_read_to_signal is a mapping of each position in the duplex
        # sequence to a signal datum from the read
        query_to_signal = read.query_to_signal
        duplex_to_read_signal = DC.map_ref_to_signal(
            query_to_signal=query_to_signal,
            ref_to_query_knots=simplex_duplex_mapping.duplex_to_simplex_mapping,
        )
        read.seq = simplex_duplex_mapping.trimmed_duplex_seq
        read.query_to_signal = duplex_to_read_signal

        read.ref_seq = None
        read.ref_to_signal = None
        read.ref_reg = None
        return read, simplex_duplex_mapping.duplex_offset

    def add_alignment(
        self, alignment_record, parse_ref_align=True, reverse_signal=False
    ):
        """Add alignment to read object

        Args:
            alignment_record (pysam.AlignedSegment)
            parse_ref_align (bool): Should reference alignment be parsed
        """
        if (
            alignment_record.reference_name is None
            and alignment_record.is_reverse
        ):
            raise RemoraError("Unmapped reads cannot map to reverse strand.")
        if self.dacs is None:
            raise RemoraError("Must add signal to io.Read before alignment.")
        self.full_align = alignment_record.to_dict()

        tags = dict(alignment_record.tags)
        try:
            self.num_trimmed = tags["ts"]
            self.dacs = self.dacs[self.num_trimmed :]
        except KeyError:
            self.num_trimmed = 0

        try:
            self.query_to_signal, self.mv_table, self.stride = parse_move_tag(
                tags["mv"],
                sig_len=self.dacs.size,
                seq_len=len(alignment_record.query_sequence),
                reverse_signal=reverse_signal,
            )
        except KeyError:
            self.query_to_signal = self.mv_table = self.stride = None

        try:
            self.shift_pa_to_norm = tags["sm"]
            self.scale_pa_to_norm = tags["sd"]
        except KeyError:
            self.compute_pa_to_norm_scaling()

        self.shift_dacs_to_norm = (
            self.shift_pa_to_norm / self.scale_dacs_to_pa
        ) - self.shift_dacs_to_pa
        self.scale_dacs_to_norm = self.scale_pa_to_norm / self.scale_dacs_to_pa

        self.seq = alignment_record.query_sequence
        if alignment_record.is_reverse:
            self.seq = util.revcomp(self.seq)
        if not parse_ref_align:
            return

        self.ref_reg = RefRegion(
            ctg=alignment_record.reference_name,
            strand="-" if alignment_record.is_reverse else "+",
            start=alignment_record.reference_start,
        )
        try:
            self.ref_seq = alignment_record.get_reference_sequence().upper()
        except ValueError:
            LOGGER.debug(
                "Reference sequence requested, but could not be extracted. "
                "Do reads contain MD tags?"
            )
            self.ref_seq = None
        self.cigar = alignment_record.cigartuples
        if alignment_record.is_reverse:
            if self.ref_seq is not None:
                self.ref_seq = util.revcomp(self.ref_seq)
            self.cigar = self.cigar[::-1]
        if self.ref_reg.ctg is not None and self.ref_seq is not None:
            self.ref_to_signal = DC.compute_ref_to_signal(
                query_to_signal=self.query_to_signal,
                cigar=self.cigar,
            )
            # +1 because knots include the end position of the last base
            assert self.ref_to_signal.size == len(self.ref_seq) + 1, (
                "discordant ref seq lengths: move+cigar:"
                f"{self.ref_to_signal.size} ref_seq:{len(self.ref_seq)}"
            )
            self.ref_reg.end = self.ref_reg.start + self.ref_to_signal.size - 1

    @classmethod
    def from_pod5_and_alignment(
        cls, pod5_read_record, alignment_record, reverse_signal=False
    ):
        """Initialize read from pod5 and pysam records

        Args:
            pod5_read_record (pod5.ReadRecord)
            alignment_record (pysam.AlignedSegment)
        """
        dacs = pod5_read_record.signal
        if reverse_signal:
            dacs = dacs[::-1]
        read = Read(
            read_id=str(pod5_read_record.read_id),
            dacs=dacs,
            shift_dacs_to_pa=pod5_read_record.calibration.offset,
            scale_dacs_to_pa=pod5_read_record.calibration.scale,
        )
        read.add_alignment(alignment_record, reverse_signal=reverse_signal)
        return read

    def into_remora_read(self, use_reference_anchor):
        """Extract RemoraRead object from io.Read.

        Args:
            use_reference_anchor (bool): Should remora read be reference
                anchored? Or basecall anchored?
        """
        if use_reference_anchor:
            if self.ref_to_signal is None:
                if self.cigar is None or self.ref_seq is None:
                    raise RemoraError("missing reference alignment")
                self.ref_to_signal = DC.compute_ref_to_signal(
                    self.query_to_signal,
                    self.cigar,
                )
                assert self.ref_to_signal.size == len(self.ref_seq) + 1, (
                    "discordant ref seq lengths: move+cigar:"
                    f"{self.ref_to_signal.size} ref_seq:{len(self.ref_seq)}"
                )

            trim_dacs = self.dacs[
                self.ref_to_signal[0] : self.ref_to_signal[-1]
            ]
            shift_seq_to_sig = self.ref_to_signal - self.ref_to_signal[0]
            seq = self.ref_seq
        else:
            trim_dacs = self.dacs[
                self.query_to_signal[0] : self.query_to_signal[-1]
            ]
            shift_seq_to_sig = self.query_to_signal - self.query_to_signal[0]
            seq = self.seq
        remora_read = DC.RemoraRead(
            dacs=trim_dacs,
            shift=self.shift_dacs_to_norm,
            scale=self.scale_dacs_to_norm,
            seq_to_sig_map=shift_seq_to_sig,
            str_seq=seq,
            read_id=self.read_id,
        )
        remora_read.check()
        return remora_read

    def set_refine_signal_mapping(self, sig_map_refiner, ref_mapping=False):
        """Refine signal mapping

        Args:
            sig_map_refiner (refine_signal_map.SigMapRefiner): Signal mapping
                refiner
            ref_mapping (bool): Refine reference mapping? Default: False/refine
                basecall to signal mapping

        Note that shift/scale parameters are overwritten for the reference or
        basecall mapping. To keep both use io.Read.copy().
        """
        if sig_map_refiner is None:
            return
        remora_read = self.into_remora_read(ref_mapping)
        remora_read.refine_signal_mapping(sig_map_refiner)
        if ref_mapping:
            self.ref_to_signal = (
                remora_read.seq_to_sig_map + self.ref_to_signal[0]
            )
        else:
            self.query_to_signal = (
                remora_read.seq_to_sig_map + self.query_to_signal[0]
            )

        self.shift_dacs_to_norm = remora_read.shift
        self.scale_dacs_to_norm = remora_read.scale
        self.shift_pa_to_norm = (
            self.shift_dacs_to_norm + self.shift_dacs_to_pa
        ) * self.scale_dacs_to_pa
        self.scale_pa_to_norm = self.scale_dacs_to_norm * self.scale_dacs_to_pa

    def get_filtered_focus_positions(self, select_focus_positions):
        """
        Args:
            select_focus_positions (dict): lookup table of (contig, strand)
            tuples (both strings) to a set of positions to include.

        Returns:
            np.ndarray of positions covered by the read and within the
            selected focus position
        """
        if self.ref_reg is None or self.ref_seq is None:
            raise RemoraError("Cannot extract focus positions without mapping")
        ref_reg = self.ref_reg
        ref_len = len(self.ref_seq)
        try:
            cs_focus_pos = select_focus_positions[(ref_reg.ctg, ref_reg.strand)]
        except KeyError:
            # no focus positions on contig/strand
            return np.array([], dtype=int)

        read_focus_ref_reg = np.array(
            sorted(
                set(range(ref_reg.start, ref_reg.start + ref_len)).intersection(
                    cs_focus_pos
                )
            ),
            dtype=int,
        )
        return (
            read_focus_ref_reg - ref_reg.start
            if ref_reg.strand == "+"
            else ref_reg.start + ref_len - read_focus_ref_reg[::-1] - 1
        )

    def get_basecall_anchored_focus_bases(
        self, motifs, select_focus_reference_positions
    ):
        """Get basecall anchored focus bases

        Args:
            motifs (list of util.Motif): List of motifs
            select_focus_reference_positions (dict): Reference focus positions
                as returned from io.parse_bed
        """
        if self.cigar is None:
            raise RemoraError("missing alignment")

        basecall_int_seq = util.seq_to_int(self.seq)
        reference_int_seq = util.seq_to_int(self.ref_seq)

        all_base_call_focus_positions = util.find_focus_bases_in_int_sequence(
            int_seq=basecall_int_seq, motifs=motifs
        )
        # mapping of reference sequence positions to base call sequence
        # positions
        mapping = DC.make_sequence_coordinate_mapping(self.cigar).astype(int)

        reference_motif_positions = (
            util.find_focus_bases_in_int_sequence(reference_int_seq, motifs)
            if select_focus_reference_positions is None
            else self.get_filtered_focus_positions(
                select_focus_positions=select_focus_reference_positions,
            )
        )
        reference_supported_focus_bases = mapping[reference_motif_positions]
        base_call_focus_bases = np.array(
            [
                focus_base
                for focus_base in all_base_call_focus_positions
                if focus_base in reference_supported_focus_bases
            ]
        )
        return base_call_focus_bases

    def copy(self):
        return deepcopy(self)

    def extract_basecall_region(self, start_base=None, end_base=None):
        """Extract region of read from basecall coordinates.

        Args:
            start_base (int): Start coordinate for region
            end_base (int): End coordinate for region

        Returns:
            ReadBasecallRegion object
        """
        start_base = start_base or 0
        end_base = end_base or self.seq_len
        reg_seq_to_sig = self.query_to_signal[start_base : end_base + 1].copy()
        reg_sig = self.norm_signal[reg_seq_to_sig[0] : reg_seq_to_sig[-1]]
        reg_seq_to_sig -= reg_seq_to_sig[0]
        return ReadBasecallRegion(
            read_id=self.read_id,
            norm_signal=reg_sig,
            seq=self.seq[start_base:end_base],
            seq_to_sig_map=reg_seq_to_sig,
            start=start_base,
        )

    def extract_ref_reg(self, ref_reg):
        """Extract region of read from reference coordinates.

        Args:
            ref_reg (RefRegion): Reference region

        Returns:
            ReadRefReg object
        """
        if self.ref_to_signal is None:
            raise RemoraError(
                "Cannot extract reference region from unmapped read"
            )
        if ref_reg.start >= self.ref_reg.start + self.ref_seq_len:
            raise RemoraError("Reference region starts after read ends")
        if ref_reg.end < self.ref_reg.start:
            raise RemoraError("Reference region ends before read starts")

        if self.ref_reg.strand == "+":
            reg_st_within_read = max(0, ref_reg.start - self.ref_reg.start)
            reg_en_within_read = ref_reg.end - self.ref_reg.start
        else:
            reg_st_within_read = max(0, self.ref_reg.end - ref_reg.end)
            reg_en_within_read = self.ref_reg.end - ref_reg.start
        reg_seq_to_sig = self.ref_to_signal[
            reg_st_within_read : reg_en_within_read + 1
        ].copy()
        reg_sig = self.norm_signal[reg_seq_to_sig[0] : reg_seq_to_sig[-1]]
        reg_seq = self.ref_seq[reg_st_within_read:reg_en_within_read]
        reg_seq_to_sig -= reg_seq_to_sig[0]
        read_reg_ref_st = max(self.ref_reg.start, ref_reg.start)
        # orient reverse strand reads on the reference
        if self.ref_reg.strand == "-":
            reg_sig = reg_sig[::-1]
            reg_seq = reg_seq[::-1]
            reg_seq_to_sig = reg_seq_to_sig[-1] - reg_seq_to_sig[::-1]
        return ReadRefReg(
            read_id=self.read_id,
            norm_signal=reg_sig,
            seq=reg_seq,
            seq_to_sig_map=reg_seq_to_sig,
            ref_reg=RefRegion(
                self.ref_reg.ctg,
                self.ref_reg.strand,
                read_reg_ref_st,
                read_reg_ref_st + len(reg_seq),
            ),
        )

    def compute_per_base_metric(
        self,
        metric=None,
        metric_func=None,
        ref_anchored=True,
        region=None,
        signal_type="norm",
        **kwargs,
    ):
        """Compute a per-base metric from a read.

        Args:
            metric (str): Named metric (e.g. dwell, mean, sd). Should be a key
                in metrics.METRIC_FUNCS
            metric_func (Callable): Function taking three arguments sequence,
                signal and a sequence to signal mapping and return a dict of
                metric names to per base metric values.
            ref_anchored (bool): Compute metric against reference bases. If
                False, return basecall anchored metrics.
            region (RefRegion): Reference region from which to extract metrics
                of bases.
            signal_type (str): Type of signal. Should be one of: norm, pa, and
                dac
            **kwargs: Extra args to pass through to metric computations

        Returns:
            Result of metric_func. Preset metrics will return dict with metric
            name keys and a numpy array of per-base metric values.
        """
        if metric is not None:
            metric_func = METRIC_FUNCS[metric]
        assert (
            metric_func is not None
        ), "Must provide either metric or metric_func"
        st_buf = en_buf = 0
        if region is None:
            seq_to_sig = (
                self.ref_to_signal if ref_anchored else self.query_to_signal
            )
        else:
            if ref_anchored:
                if (
                    self.ref_reg.ctg != region.ctg
                    or self.ref_reg.strand != region.strand
                ):
                    raise RemoraError("Region contig/strand do not match read")
                if (
                    region.start >= self.ref_reg.end
                    or self.ref_reg.start >= region.end
                ):
                    raise RemoraError("Region does not overlap read.")
                if self.ref_reg.strand == "+":
                    st_coord = region.start - self.ref_reg.start
                    en_coord = region.end - self.ref_reg.start
                else:
                    st_coord = self.ref_reg.end - region.end
                    en_coord = self.ref_reg.end - region.start
                if st_coord < 0:
                    st_buf = -st_coord
                    st_coord = 0
                if en_coord > self.ref_seq_len:
                    en_buf = en_coord - self.ref_seq_len
                    en_coord = self.ref_seq_len
                seq_to_sig = self.ref_to_signal[st_coord : en_coord + 1]
            else:
                if region.start < 0 or region.start > self.seq_len:
                    raise RemoraError("Region does not overlap read.")
                # TODO deal with partially overlapping region
                st_buf = en_buf = 0
                seq_to_sig = self.query_to_signal[region.start : region.end]
        if signal_type == "norm":
            sig = self.norm_signal
        elif signal_type == "pa":
            sig = self.pa_signal
        elif signal_type == "dac":
            sig = self.dacs
        else:
            assert False, "signal_type must be norm, pa or dac"

        metrics_vals = metric_func(sig, seq_to_sig, **kwargs)
        if max(st_buf, en_buf) > 0:
            tmp_metrics_vals = {}
            for metric_name, metric_vals in metrics_vals.items():
                tmp_metrics_vals[metric_name] = np.full(region.len, np.NAN)
                tmp_metrics_vals[metric_name][
                    st_buf : st_buf + metric_vals.size
                ] = metric_vals
            metrics_vals = tmp_metrics_vals
        return metrics_vals


##########
# Duplex #
##########


@dataclass
class DuplexRead:
    duplex_read_id: str
    duplex_alignment: dict
    is_reverse_mapped: bool
    template_read: Read
    complement_read: Read
    template_ref_start: int
    complement_ref_start: int

    @classmethod
    def from_reads_and_alignment(
        cls,
        *,
        template_read: Read,
        complement_read: Read,
        duplex_alignment: AlignedSegment,
    ):
        is_reverse_mapped = duplex_alignment.is_reverse
        duplex_direction_read, reverse_complement_read = (
            (template_read, complement_read)
            if not is_reverse_mapped
            else (complement_read, template_read)
        )

        (
            template_read,
            template_ref_start,
        ) = duplex_direction_read.with_duplex_alignment(
            duplex_alignment,
            duplex_orientation=True,
        )
        (
            complement_read,
            complement_ref_start,
        ) = reverse_complement_read.with_duplex_alignment(
            duplex_alignment,
            duplex_orientation=False,
        )

        return DuplexRead(
            duplex_read_id=duplex_alignment.query_name,
            duplex_alignment=duplex_alignment.to_dict(),
            is_reverse_mapped=is_reverse_mapped,
            template_read=template_read,
            complement_read=complement_read,
            template_ref_start=template_ref_start,
            complement_ref_start=complement_ref_start,
        )

    @property
    def duplex_basecalled_sequence(self):
        # n.b. pysam reverse-complements the query sequence on reverse mappings
        # [https://pysam.readthedocs.io/en/latest/api.html#pysam.AlignedSegment.query_sequence]
        return self.duplex_alignment["seq"]


class DuplexPairsBuilder:
    def __init__(self, simplex_index, pod5_path):
        """Duplex Pairs Builder

        Args:
            simplex_index (ReadIndexedBam): Simplex bam index
            pod5_path (str): Path to POD5 file
        """
        self.simplex_index = simplex_index
        self.pod5_path = pod5_path
        self.reader = pod5.Reader(Path(pod5_path))

    def make_read_pair(self, read_id_pair):
        """Make read pair

        Args:
            read_id_pair (tuple): 2-tuple of read ID strings
        """
        try:
            pod5_reads = list(
                self.reader.reads(
                    selection=list(read_id_pair), preload=["samples"]
                )
            )
        except RuntimeError:
            return None, "duplex pair read id(s) missing from pod5"

        if len(pod5_reads) < 2:
            return None, "duplex pair read id(s) missing from pod5"
        if len(pod5_reads) > 2:
            return None, "pod5 has multiple reads with the same id"
        pod5_reads = {str(read.read_id): read for read in pod5_reads}

        temp_read_id, comp_read_id = read_id_pair
        try:
            temp_align = self.simplex_index.get_first_alignment(temp_read_id)
            comp_align = self.simplex_index.get_first_alignment(comp_read_id)
        except RemoraError:
            return None, "failed to find read in simplex bam"
        temp_io_read = Read.from_pod5_and_alignment(
            pod5_read_record=pod5_reads[temp_read_id],
            alignment_record=temp_align,
        )
        comp_io_read = Read.from_pod5_and_alignment(
            pod5_read_record=pod5_reads[comp_read_id],
            alignment_record=comp_align,
        )

        return (temp_io_read, comp_io_read), None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.reader.close()
        self.simplex_index.close()



import sys
from setuptools import setup, Extension


extra_compile_args = ["-std=c99"]
if sys.platform == "darwin":
    extra_compile_args.append("-mmacosx-version-min=10.9")
    print("Using macOS clang args")
ext_modules = [
    Extension(
        "remora.encoded_kmers",
        sources=["src/remora/encoded_kmers.pyx"],
        extra_compile_args=extra_compile_args,
        language="c",
    ),
    Extension(
        "remora.refine_signal_map_core",
        sources=["src/remora/refine_signal_map_core.pyx"],
        extra_compile_args=extra_compile_args,
        language="c",
    ),
]


if __name__ == "__main__":
    setup(
        use_pyscaffold=True,
        setup_requires=[
            "setuptools>=38.3",
            "cython",
        ],
        ext_modules=ext_modules,
    )
